/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.authservice;

import java.util.Map;

import org.osgi.annotation.versioning.ProviderType;

/**
 * Denotes the service that is responsible for keeping track which users are logged in and out.
 */
@ProviderType
public interface AuthenticationService {

    /**
     * Called when a user is attempting to log in to the application.
     * <p>
     * If the log in fails, an {@link AuthenticationException} should be thrown to indicate
     * this. If no such exception is thrown, the login is assumed to be successful.
     * </p>
     * <p>
     * This method accepts an optional context for this login, which can be provided and
     * filled by the caller to supply additional information about this action. For example,
     * the context can contain the location/IP address of the user, user agent information,
     * and so on.
     * </p>
     *
     * @param credentials the credentials that identify the (local) user, cannot be <code>null</code>;
     * @param context any (optional) additional context from the user that can be helpful in the
     *        login process, can be <code>null</code>.
     */
    void login(Map<String, String> credentials, Map<String, String> context) throws AuthenticationException;

    /**
     * Tests whether or not users are currently logged in.
     *
     * @param credentials the credentials that identify the (local) user, cannot be <code>null</code>.
     * @return <code>true</code> if the user is currently logged in, <code>false</code> otherwise.
     */
    boolean isLoggedIn(Map<String, String> credentials);

    /**
     * Called when a user is logging out of the application.
     * <p>
     * If the log out fails, an {@link AuthenticationException} should be thrown to indicate
     * this. If no such exception is thrown, the logout is assumed to be successful.
     * </p>
     *
     * @param credentials the credentials that identify the (local) user, cannot be <code>null</code>.
     */
    void logout(Map<String, String> credentials) throws AuthenticationException;
}
