/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.authservice;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import javax.servlet.http.HttpServletResponse;

/**
 * Provides information about why an {@link AuthenticationHandler} rejected a particular request.
 */
public class RejectInfo {
    private final String m_statusCode;
    private final Throwable m_cause;
    private final URI m_failureURL;
    private final Map<String, List<String>> m_headers;

    /**
     * Creates a new {@link RejectInfo} instance.
     */
    public RejectInfo(String statusCode, URI failureURL) {
        this(statusCode, failureURL, null);
    }

    /**
     * Creates a new {@link RejectInfo} instance.
     */
    public RejectInfo(String statusCode, URI failureURL, Throwable cause) {
        m_statusCode = statusCode;
        m_failureURL = failureURL;
        m_cause = cause;
        m_headers = new HashMap<>();
    }

    public static Optional<RejectInfo> create(String statusCode, URI failureURL) {
        return Optional.of(new RejectInfo(statusCode, failureURL, null));
    }

    public static Optional<RejectInfo> create(String statusCode, URI failureURL, Throwable cause) {
        return Optional.of(new RejectInfo(statusCode, failureURL, cause));
    }

    public static Optional<RejectInfo> none() {
        return Optional.empty();
    }

    /**
     * Applies this {@link RejectInfo} to the given {@link HttpServletResponse} redirecting
     * the client to the contained failure URL.
     */
    public void applyTo(HttpServletResponse resp) throws IOException {
        applyTo(resp, true /* redirectToFailureURL */);
    }

    /**
     * Applies this {@link RejectInfo} to the given {@link HttpServletResponse} optionally
     * redirecting the client to the contained failure URL.
     */
    public void applyTo(HttpServletResponse resp, boolean redirectToFailureURL) throws IOException {
        m_headers.entrySet().stream().forEach(outer -> {
            outer.getValue().stream().forEach(value -> {
                resp.setHeader(outer.getKey(), value);
            });
        });

        resp.setHeader("X-AuthFailure-Code", m_statusCode);

        if (m_cause != null) {
            resp.setHeader("X-AuthFailure-Cause", m_cause.getMessage());
        }

        if (m_failureURL != null && redirectToFailureURL) {
            resp.sendRedirect(m_failureURL.toASCIIString());
        }
    }

    /**
     * Adds an header that should be send with a servlet response.
     *
     * @param name the name of the header;
     * @param value the value of the header.
     */
    public void addHeader(String name, String value) {
        m_headers.computeIfAbsent(name, ignored -> new ArrayList<>()).add(value);
    }

    /**
     * @return a {@link Throwable} that indicates the cause of the rejection, can be <code>null</code>.
     */
    public Throwable getCause() {
        return m_cause;
    }

    /**
     * @return a URI that can be used to redirect to a proper landing page for the user.
     */
    public URI getFailureURL() {
        return m_failureURL;
    }

    /**
     * @return the (optional) response headers to send back, cannot be <code>null</code>.
     */
    public Map<String, List<String>> getHeaders() {
        Map<String, List<String>> result = new HashMap<>();
        for (Map.Entry<String, List<String>> entry : m_headers.entrySet()) {
            result.put(entry.getKey(), new ArrayList<>(entry.getValue()));
        }
        return result;
    }

    /**
     * @return the status code that indicates the cause of failure, should not be <code>null</code>.
     */
    public String getStatusCode() {
        return m_statusCode;
    }
}
