/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.authservice.impl;

import static org.amdatu.security.authentication.authservice.AuthenticationServiceConstants.KEY_PROVIDER_TYPE;
import static org.amdatu.security.authentication.authservice.AuthenticationServiceConstants.KEY_SESSION_ID;
import static org.amdatu.security.authentication.authservice.AuthenticationServiceConstants.KEY_USER_ID;
import static org.amdatu.security.authentication.authservice.AuthenticationServiceConstants.TOPIC_LOGIN;
import static org.amdatu.security.authentication.authservice.AuthenticationServiceConstants.TOPIC_LOGOUT;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.PROVIDER_TYPE;
import static org.amdatu.security.tokenprovider.TokenConstants.SUBJECT;

import java.util.HashMap;
import java.util.Map;

import org.amdatu.security.authentication.authservice.AuthenticationException;
import org.amdatu.security.authentication.authservice.AuthenticationService;
import org.amdatu.security.authentication.authservice.SessionStore;
import org.osgi.service.event.Event;
import org.osgi.service.event.EventAdmin;
import org.osgi.service.log.LogService;

/**
 * Default implementation of {@link AuthenticationService}, which uses the subject
 * contained in credentials and maps this to a session identifier.
 */
public class AuthenticationServiceImpl implements AuthenticationService {
    private static final String SESSION_ID = "sessionID";

    private static final String ERR_ACCESS_DENIED = "access_denied";

    // Managed by Felix DM...
    private volatile SessionStore m_sessionStore;
    private volatile EventAdmin m_eventAdmin;
    private volatile LogService m_log;

    @Override
    public boolean isLoggedIn(Map<String, String> credentials) {
        String userId = credentials.get(SUBJECT);
        String sessionId = credentials.get(SESSION_ID);

        if (userId == null || sessionId == null) {
            // Don't bother to check something we can see right away...
            return false;
        }

        return m_sessionStore.containsSession(userId, sessionId);
    }

    @Override
    public void login(Map<String, String> credentials, Map<String, String> context) throws AuthenticationException {
        String userId = credentials.get(SUBJECT);
        String providerType = credentials.get(PROVIDER_TYPE);

        // Generate a new session ID...
        String sessionId = m_sessionStore.createSession(userId, context);

        credentials.put(SESSION_ID, sessionId);

        postEvent(TOPIC_LOGIN, context, KEY_USER_ID, userId, KEY_SESSION_ID, sessionId, KEY_PROVIDER_TYPE, providerType);

        m_log.log(LogService.LOG_INFO, "Login successful for " + userId + "! Session ID is " + sessionId);
    }

    @Override
    public void logout(Map<String, String> credentials) {
        if (credentials.isEmpty()) {
            // Nothing to do...
            return;
        }

        String userId = credentials.get(SUBJECT);
        String sessionId = credentials.get(SESSION_ID);
        String providerType = credentials.get(PROVIDER_TYPE);

        if (!m_sessionStore.closeSession(userId, sessionId)) {
            m_log.log(LogService.LOG_WARNING, "Logout failed for " + credentials + ": invalid session ID!");
            throw new AuthenticationException(ERR_ACCESS_DENIED);
        }

        postEvent(TOPIC_LOGOUT, null, KEY_USER_ID, userId, KEY_SESSION_ID, sessionId, KEY_PROVIDER_TYPE, providerType);

        m_log.log(LogService.LOG_INFO, "Logout succeeded for " + userId + "! Session ID was " + sessionId);
    }

    private void postEvent(String topic, Map<String, String> context, Object... eventProps) {
        Map<String, Object> props = new HashMap<>();
        for (int i = 0; i < eventProps.length; i += 2) {
            props.put(eventProps[i].toString(), eventProps[i + 1]);
        }
        if (context != null) {
            context.entrySet().stream()
                .forEach(entry -> props.put("context.".concat(entry.getKey()), entry.getValue()));
        }
        m_eventAdmin.postEvent(new Event(topic, props));
    }
}
