/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.authservice.rest;

import static java.util.Collections.singletonMap;
import static javax.ws.rs.core.Response.ok;
import static javax.ws.rs.core.Response.seeOther;
import static javax.ws.rs.core.Response.status;
import static javax.ws.rs.core.Response.Status.FORBIDDEN;
import static javax.ws.rs.core.Response.Status.OK;
import static org.amdatu.security.authentication.authservice.rest.util.HttpUtil.SET_COOKIE_HEADER;
import static org.amdatu.security.authentication.authservice.rest.util.HttpUtil.getContext;
import static org.amdatu.security.authentication.authservice.rest.util.HttpUtil.isXmlHttpRequest;
import static org.amdatu.security.authentication.authservice.rest.util.HttpUtil.toHttpHeader;
import static org.amdatu.security.authentication.authservice.rest.util.TokenUtil.extendTokenLifetime;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_ACCESS_DENIED;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_INTERACTION_REQUIRED;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_INVALID_REQUEST;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_LOGGED_OUT;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_LOGIN_REQUIRED;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_UNAUTHORIZED_CLIENT;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.ERROR_CODE;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.PROMPT;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.PROMPT_LOGIN;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.PROVIDER_TYPE;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.ACCOUNT_ID;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.ID_TOKEN;
import static org.amdatu.security.tokenprovider.TokenConstants.SUBJECT;
import static org.amdatu.security.tokenprovider.http.TokenUtil.getTokenFromCookie;
import static org.amdatu.web.util.HttpRequestUtil.getRequestURL;
import static org.amdatu.web.util.HttpRequestUtil.isSecureRequest;
import static org.osgi.service.http.context.ServletContextHelper.REMOTE_USER;
import static org.osgi.service.log.LogService.LOG_DEBUG;
import static org.osgi.service.log.LogService.LOG_INFO;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.ws.rs.Consumes;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.NewCookie;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import javax.ws.rs.core.Response.Status;

import org.amdatu.security.authentication.authservice.AuthenticationException;
import org.amdatu.security.authentication.authservice.AuthenticationHandler;
import org.amdatu.security.authentication.authservice.AuthenticationService;
import org.amdatu.security.authentication.authservice.RejectInfo;
import org.amdatu.security.authentication.authservice.rest.util.TokenUtil.RenewalResult;
import org.amdatu.security.authentication.idprovider.IdProvider;
import org.amdatu.security.tokenprovider.InvalidTokenException;
import org.amdatu.security.tokenprovider.InvalidTokenException.Reason;
import org.amdatu.security.tokenprovider.TokenProvider;
import org.amdatu.web.rest.doc.Description;
import org.amdatu.web.util.URIBuilder;
import org.osgi.framework.ServiceReference;
import org.osgi.service.cm.ConfigurationException;
import org.osgi.service.cm.ManagedService;
import org.osgi.service.log.LogService;

/**
 * Provides a front-end REST resource that allows you to login and logout from a system.
 */
@Path("/")
public class AuthenticationResource implements ManagedService, AuthenticationHandler {
    private static class TokenProperties {
        final Optional<Reason> m_reason;
        final Map<String, String> m_userIdentity;

        public TokenProperties(Map<String, String> userIdentity, Reason reason) {
            m_userIdentity = userIdentity == null ? new HashMap<>() : userIdentity;
            m_reason = Optional.ofNullable(reason);
        }

        /**
         * @return the name of the provider (as included in the user identity).
         */
        Optional<String> getProviderType() {
            return Optional.ofNullable(m_userIdentity.get(PROVIDER_TYPE));
        }

        /**
         * @return <code>true</code> iff the token was expired.
         */
        boolean isTokenExpired() {
            return m_reason.map(r -> Reason.TOKEN_EXPIRED.equals(r)).orElse(Boolean.FALSE);
        }

        /**
         * @return <code>true</code> iff the token was valid and contains attributes that identify the user.
         */
        boolean isValid() {
            return !m_reason.isPresent() && !m_userIdentity.isEmpty();
        }
    }

    // Injected by Felix DM...
    private volatile AuthenticationService m_authService;
    private volatile TokenProvider m_tokenProvider;
    private volatile LogService m_log;

    private final Map<String, IdProvider> m_idProviders;
    // Locally managed...
    private volatile AuthenticationResourceConfig m_config;

    /**
     * Creates a new {@link AuthenticationResource} instance.
     */
    public AuthenticationResource() {
        m_idProviders = new ConcurrentHashMap<>();
        m_config = new AuthenticationResourceConfig();
    }

    @GET
    @Path("/login/{provider}")
    @Produces(MediaType.APPLICATION_JSON)
    @Description("Called by the identity provider (= local or remote) to return the result of the authentication process.")
    public Response finishLogin(@PathParam("provider") String providerType, @Context HttpServletRequest req)
        throws IOException {
        AuthenticationResourceConfig cfg = m_config;

        // AMDATUSEC-31 ensure that the callback URI does not include any query string or fragment...
        URI callbackURI = getRequestURL(req)
            .setFragment(null)
            .setQuery(null)
            .build();

        // Determine the actual user identity using the IdProvider. Note that this
        // could imply a call to an external service...
        Map<String, String> userIdentity = getProvider(providerType)
            .map(p -> p.getUserIdentity(callbackURI, req.getParameterMap()))
            .orElseGet(() -> singletonMap(ERROR_CODE, CODE_UNAUTHORIZED_CLIENT));

        String error = userIdentity.get(ERROR_CODE);

        if (CODE_INTERACTION_REQUIRED.equals(error) || CODE_LOGIN_REQUIRED.equals(error)) {
            // Force the user to re-authenticate itself...
            log(LOG_INFO, req, providerType, userIdentity, "Forcing new login: ID provider needs explicity login!");

            // Try to enforce a login at the remote IdProvider...
            Map<String, String[]> params = new HashMap<>(req.getParameterMap());
            params.put(PROMPT, new String[] { PROMPT_LOGIN });

            return getAuthenticationURI(providerType, callbackURI, params)
                .map(uri -> seeOther(uri).build())
                .orElse(respondFailure(cfg, req, error));
        }
        else if (error != null) {
            log(LOG_INFO, req, providerType, userIdentity, "Login failed: ID provider returned error code %s!", error);

            return respondFailure(cfg, req, error);
        }

        try {
            // Determine whether the user is actually known for our application...
            m_authService.login(userIdentity, getContext(req));

            log(LOG_DEBUG, req, providerType, userIdentity, "Login successful!");

            return addTokenCookie(respondSuccess(req, cfg.getLandingPageURL()),
                createTokenCookie(cfg, req, userIdentity)).build();
        }
        catch (AuthenticationException e) {
            log(LOG_INFO, req, providerType, userIdentity, "User unknown or provided invalid credentials!");

            return respondFailure(cfg, req, e.getCode());
        }
    }

    @GET
    @Path("/login/providers")
    @Produces(MediaType.APPLICATION_JSON)
    @Description("Returns the names of the available identity providers that can be used to authenticate users.")
    public List<String> getIdProviders() throws IOException {
        return new ArrayList<>(m_idProviders.keySet());
    }

    /**
     * This method is used by all users of {@link AuthenticationHandler} and is implemented here as we need access
     * to the configured failure page and cookie name.
     */
    @Override
    public Optional<RejectInfo> handleSecurity(HttpServletRequest request, HttpServletResponse response)
        throws IOException {
        AuthenticationResourceConfig cfg = m_config;

        String errorCode = CODE_ACCESS_DENIED;
        URI failureURL = cfg.getFailureURL();

        TokenProperties tokenProps = getTokenProperties(cfg, request);

        String remoteUser = null;
        String providerType = tokenProps.getProviderType().orElse(null);

        if (providerType == null) {
            // We're unable to determine what we should do here, bail out for safety...
            return RejectInfo.create(CODE_INVALID_REQUEST, failureURL);
        }

        Map<String, String> userIdentity = tokenProps.m_userIdentity;

        if (m_authService.isLoggedIn(userIdentity)) {
            if (tokenProps.isValid()) {
                remoteUser = userIdentity.get(SUBJECT);
            }
            else if (tokenProps.isTokenExpired() && tokenLifetimeExtended(cfg, userIdentity)) {
                remoteUser = userIdentity.get(SUBJECT);

                // Update the token cookie, to prolong the users' session...
                response.addHeader(SET_COOKIE_HEADER, toHttpHeader(createTokenCookie(cfg, request, userIdentity)));
            }
            else {
                // Token no longer valid and we're not allowed to auto-renew it...
                log(LOG_DEBUG, request, providerType, userIdentity,
                    "Token not valid. Ending session to force new login...");
                // Give the caller a clue what is required of him...
                errorCode = CODE_LOGIN_REQUIRED;

                m_authService.logout(userIdentity);
            }
        }

        if (remoteUser != null) {
            // User was logged in, and its token was still valid or automatically renewed...
            request.setAttribute(REMOTE_USER, remoteUser);
            request.setAttribute(PROVIDER_TYPE, providerType);

            // In case of local login put the account id on the request.
            String accountId = userIdentity.get(ACCOUNT_ID);
            if (accountId != null) {
                request.setAttribute(ACCOUNT_ID, accountId);
            }

            // In case of OpenIDConnect login put the id token on the request.
            String idToken = userIdentity.get(ID_TOKEN);
            if (accountId != null) {
                request.setAttribute(ID_TOKEN, idToken);
            }

            return RejectInfo.none();
        }

        log(LOG_DEBUG, request, providerType, userIdentity, "No/invalid token provided, or user not logged in!");
        // Clear any of the existing token cookies. This is not an optional thing...
        response.addHeader(SET_COOKIE_HEADER, toHttpHeader(createTokenCookie(cfg, request, null)));

        return RejectInfo.create(errorCode, failureURL);
    }

    @POST
    @Path("/login")
    @Consumes(MediaType.APPLICATION_FORM_URLENCODED)
    @Description("Called by the client/user agent to initiate the authentication with the identity provider.")
    public Response initiateLogin(@Context HttpServletRequest req) throws IOException {
        AuthenticationResourceConfig cfg = m_config;

        // We want any other form-parameters to retain as-is, hence we do not use the
        // FormParam annotation of JAXRS as it mangles our request parameters!
        String providerType = req.getParameter(PROVIDER_TYPE);

        TokenProperties tokenProps = getTokenProperties(cfg, req);
        Map<String, String> userIdentity = tokenProps.m_userIdentity;

        if (tokenProps.isValid()) {
            try {
                // Use the same provider as the token indicates, this way, we'll always
                // will use the right provider to refresh our token...
                providerType = tokenProps.getProviderType().orElse(providerType);

                ResponseBuilder respBldr = respondSuccess(req, cfg.getLandingPageURL());

                // Try to refresh the token, and if refreshed, update the cookie for the
                // user...
                final String pType = providerType;
                getProvider(pType)
                    .ifPresent(provider -> {
                        log(LOG_DEBUG, req, pType, userIdentity, "Trying to refresh token...");

                        if (provider.refreshToken(userIdentity)) {
                            log(LOG_DEBUG, req, pType, userIdentity, "Token refreshed.");

                            addTokenCookie(respBldr, createTokenCookie(cfg, req, userIdentity));
                        }
                        else {
                            log(LOG_DEBUG, req, pType, userIdentity, "Token NOT refreshed.");
                        }
                    });

                // In case we're already logged in (and our token is apparently still valid),
                // we can go straight through our final destination...
                return respBldr.build();
            }
            catch (InvalidTokenException e) {
                log(LOG_DEBUG, req, providerType, userIdentity, e, "Token refresh failed. Forcing new login.");
                // Fall through, and try to restart the whole login sequence...
            }
        }

        log(LOG_DEBUG, req, providerType, userIdentity, "Initiating new login...");

        // We want to get called back on this endpoint, but with a GET method instead
        // and with the provider appended. Also, ensure that the callback URI does not
        // include any query string or fragment (see AMDATUSEC-31)...
        URI callbackURI = getRequestURL(req)
            .appendToPath(providerType)
            .setFragment(null)
            .setQuery(null)
            .build();

        // See if there is a identity provider supplied as parameter, if so, get its authorization URI, otherwise return a failure...
        return getAuthenticationURI(providerType, callbackURI, req.getParameterMap())
            .map(uri -> seeOther(uri).build())
            .orElse(respondFailure(cfg, req, CODE_INVALID_REQUEST));
    }

    @GET
    @Path("/login")
    @Produces(MediaType.APPLICATION_JSON)
    @Description("Checks whether the request contains a valid token indicating that the user is properly authenticated.")
    public Response isLoggedIn(@Context HttpServletRequest req) {
        AuthenticationResourceConfig cfg = m_config;

        TokenProperties props = getTokenProperties(cfg, req);

        if (!props.isValid() || !m_authService.isLoggedIn(props.m_userIdentity)) {
            // No or invalid token, not authenticated...
            return status(FORBIDDEN).build();
        }

        return status(OK).entity(props.m_userIdentity).build();
    }

    @POST
    @Path("/logout")
    @Produces(MediaType.APPLICATION_JSON)
    @Description("Called by the client/user agent to log out.")
    public Response logout(@Context HttpServletRequest req) throws IOException {
        AuthenticationResourceConfig cfg = m_config;

        TokenProperties tokenProps = getTokenProperties(cfg, req);
        Map<String, String> userIdentity = tokenProps.m_userIdentity;

        String providerId = userIdentity.get(PROVIDER_TYPE);
        String code = CODE_LOGGED_OUT;
        URI destURI = cfg.getLoginURL();
        URI callbackURI = destURI;

        try {
            m_authService.logout(userIdentity);

            if (cfg.isEndSessionAtLogout()) {
                log(LOG_INFO, req, providerId, userIdentity, "Ending session at ID provider...");

                destURI = getProvider(providerId)
                    .flatMap(p -> p.getEndSessionURI(callbackURI, userIdentity))
                    .orElse(createURI(destURI, code));
            }
        }
        catch (AuthenticationException e) {
            log(LOG_INFO, req, providerId, userIdentity, e, "Log out failed!");
            code = e.getCode();
        }

        return addTokenCookie(seeOther(createURI(destURI, code)), createTokenCookie(cfg, req, null)).build();
    }

    @Override
    public void updated(Dictionary<String, ?> config) throws ConfigurationException {
        AuthenticationResourceConfig newConfig;
        if (config != null) {
            newConfig = new AuthenticationResourceConfig(config);
        }
        else {
            newConfig = new AuthenticationResourceConfig();
        }
        m_config = newConfig;
    }

    /**
     * Called by Felix DM for all {@link IdProvider}s that are registered.
     */
    protected final void addIdProvider(ServiceReference<IdProvider> srvRef, IdProvider srv) {
        String providerType = (String) srvRef.getProperty(PROVIDER_TYPE);
        m_idProviders.put(providerType, srv);
    }

    /**
     * Called by Felix DM for all {@link IdProvider}s that are registered.
     */
    protected final void removeIdProvider(ServiceReference<IdProvider> srvRef, IdProvider srv) {
        String providerType = (String) srvRef.getProperty(PROVIDER_TYPE);
        m_idProviders.remove(providerType);
    }

    /**
     * Adds a HTTP header to the given response builder with the contents of the given cookie.
     * <p>
     * The reason for using this method is that this one properly handles certain cookie flags that
     * are not supported by the JAX-RS API, such as http-only cookies.
     * </p>
     *
     * @param builder the response builder to add the cookie header to, cannot be <code>null</code>;
     * @param cookie the (JAX-RS) cookie to add, cannot be <code>null</code>.
     * @return the given responsebuilder, never <code>null</code>.
     */
    private ResponseBuilder addTokenCookie(ResponseBuilder builder, NewCookie cookie) {
        return builder.header(SET_COOKIE_HEADER, toHttpHeader(cookie));
    }

    /**
     * @return a cookie instance. Do <b>not</b> use this cookie directly in the JAX-RS API, see
     *         {@link #addTokenCookie(ResponseBuilder, NewCookie)} for the rationale behind this!
     */
    private NewCookie createTokenCookie(AuthenticationResourceConfig cfg, HttpServletRequest req,
        Map<String, String> tokenAttrs) {
        boolean secure = isSecureRequest(req);
        String token = null;
        int maxAge = 0; // will cause our cookie to be deleted right away...

        if (tokenAttrs != null && !tokenAttrs.isEmpty()) {
            token = m_tokenProvider.generateToken(tokenAttrs);
            maxAge = cfg.getCookieMaxAge();
        }

        return new NewCookie(cfg.getCookieName(), token, cfg.getCookiePath(), cfg.getCookieDomain(), null, maxAge,
            secure);
    }

    private URI createURI(URI uri, String code) {
        URIBuilder builder = new URIBuilder(uri);
        if (code != null) {
            builder.appendToQuery("msg", code);
        }
        return builder.build();
    }

    private Optional<URI> getAuthenticationURI(String providerId, URI callbackURI, Map<String, String[]> params) {
        return getProvider(providerId).map(provider -> provider.getAuthenticationURI(callbackURI, params));
    }

    private Optional<IdProvider> getProvider(String id) {
        return Optional.ofNullable(m_idProviders.get(id));
    }

    private TokenProperties getTokenProperties(AuthenticationResourceConfig cfg, HttpServletRequest req) {
        Map<String, String> properties = null;
        Reason reason = null;

        try {
            String token = getTokenFromCookie(req, cfg.getCookieName());
            properties = m_tokenProvider.verifyToken(token);
        }
        catch (IllegalArgumentException e) {
            // Token wasn't supplied at all...
            reason = Reason.UNKNOWN;
        }
        catch (InvalidTokenException e) {
            // Token has expired or wasn't supplied at all...
            reason = e.getReason();
            properties = e.getAttributes().orElse(null);
        }

        return new TokenProperties(properties, reason);
    }

    private void log(int level, HttpServletRequest req, String provider, Map<String, String> userIdentity, String msg,
        Object... args) {
        log(level, req, provider, userIdentity, null, msg, args);
    }

    private void log(int level, HttpServletRequest req, String provider, Map<String, String> userIdentity, Throwable t,
        String msg, Object... args) {
        StringBuilder sb = new StringBuilder();
        if (userIdentity != null) {
            sb.append(" [user identity: ").append(userIdentity.get(SUBJECT)).append("]");
        }
        if (provider != null) {
            sb.append(" [provider: ").append(provider).append("]");
        }
        if (req != null) {
            sb.append(" [request origin: ").append(req.getRemoteAddr()).append(":").append(req.getRemotePort())
                .append("]");
            sb.append(" [requested url: ").append(req.getRequestURL()).append("]");
        }
        sb.append(" - ").append(String.format(msg, args));

        if (t != null) {
            m_log.log(level, sb.toString());
        }
        else {
            m_log.log(level, sb.toString(), t);
        }
    }

    private Response respondFailure(AuthenticationResourceConfig cfg, HttpServletRequest req, String reason)
        throws IOException {
        URIBuilder builder = new URIBuilder(cfg.getFailureURL());
        if (reason != null) {
            builder.appendToQuery("msg", reason);
        }
        URI destURI = builder.build();

        ResponseBuilder rb;
        if (isXmlHttpRequest(req) || destURI == null) {
            // Ajax request or no failure page: respond with a forbidden
            // and a plain body instead of redirecting to a failure page
            // (this reduced traffic a bit as the client probably isn't
            // interested in the redirected page but more in the result
            // of the operation)...
            rb = status(Status.FORBIDDEN).entity(reason);
        }
        else {
            rb = seeOther(destURI);
        }
        return addTokenCookie(rb, createTokenCookie(cfg, req, null)).build();
    }

    private ResponseBuilder respondSuccess(HttpServletRequest req, URI destURI) {
        ResponseBuilder rb;
        if (isXmlHttpRequest(req) || destURI == null) {
            // Ajax request or no success page: respond with an empty OK.
            // (this reduced traffic a bit as the client probably isn't
            // interested in the redirected page but more in the result
            // of the operation)...
            rb = ok();
        }
        else {
            rb = seeOther(destURI);
        }
        return rb;
    }

    private boolean tokenLifetimeExtended(AuthenticationResourceConfig cfg, Map<String, String> userIdentity) {
        if (!cfg.isAutoRenewToken()) {
            return false;
        }

        String subject = userIdentity.get(SUBJECT);

        RenewalResult result = extendTokenLifetime(userIdentity, cfg.getTokenMaxLifetime());
        if (result == RenewalResult.MAX_LIFETIME_REACHED) {
            // Token is renewed too often...
            m_log.log(LOG_INFO,
                String.format("Not auto-renewing token for %s: maximum token lifetime reached!", subject));

            return false;
        }
        else if (result == RenewalResult.STILL_VALID) {
            // Token is still valid, so nothing needs to be renewed...
            m_log.log(LOG_DEBUG, String.format("Not auto-renewing token for %s: token is still valid!", subject));

            return false;
        }

        m_log.log(LOG_DEBUG,
            String.format("Token renewed for %s: token was no longer valid, but could be renewed!", subject));

        // Token was renewed...
        return true;
    }
}
