/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.authservice.rest;

import static org.amdatu.security.authentication.authservice.rest.util.ConfigUtil.getBoolean;
import static org.amdatu.security.authentication.authservice.rest.util.ConfigUtil.getInteger;
import static org.amdatu.security.authentication.authservice.rest.util.ConfigUtil.getLong;
import static org.amdatu.security.authentication.authservice.rest.util.ConfigUtil.getString;
import static org.amdatu.security.authentication.authservice.rest.util.ConfigUtil.getURI;

import java.net.URI;
import java.time.Duration;
import java.util.Dictionary;

import org.amdatu.security.tokenprovider.http.TokenUtil;
import org.osgi.service.cm.ConfigurationException;

/**
 * Provides the configuration for {@link AuthenticationResource}.
 */
public class AuthenticationResourceConfig {
    public static final String KEY_LANDING_PAGE_URL = "landingPageURL";
    public static final String KEY_LOGIN_URL = "loginURL";
    public static final String KEY_FAILURE_URL = "failureURL";
    public static final String KEY_COOKIE_NAME = "cookieName";
    public static final String KEY_COOKIE_DOMAIN = "cookieDomain";
    public static final String KEY_COOKIE_PATH = "cookiePath";
    public static final String KEY_COOKIE_MAX_AGE = "cookieMaxAge";
    /**
     * OPTIONAL. Defines whether or not the "end session" endpoint of the (remote) identity provider should be called when the user logs out. Defaults to <code>true</code>.
     */
    public static final String KEY_END_SESSION_AT_LOGOUT = "endSessionAtLogout";
    /**
     * OPTIONAL. Defines whether or not expired token will be auto-renewed in case the user is still logged in.
     */
    public static final String KEY_AUTO_RENEW_TOKEN = "autoRenewToken";
    /**
     * OPTIONAL. Defines what the maximum lifetime (in seconds) of a token can be in case tokens are auto renewed. The maximum lifetime is determined on the issuer time of the token.
     */
    public static final String KEY_TOKEN_MAX_LIFETIME = "tokenMaxLifetime";

    private static final String DEFAULT_TOKEN_COOKIE_NAME = TokenUtil.AMDATU_TOKEN_COOKIE_NAME;
    private static final String DEFAULT_COOKIE_DOMAIN = null;
    private static final String DEFAULT_COOKIE_PATH = "/";
    private static final int DEFAULT_COOKIE_MAX_AGE = -1;
    private static final String DEFAULT_URI = "../index.html";
    private static final boolean DEFAULT_END_SESSION_AT_LOGOUT = true;
    private static final boolean DEFAULT_AUTO_RENEW_TOKEN = false;
    private static final long DEFAULT_TOKEN_MAX_LIFETIME = Duration.ofDays(7).getSeconds();

    private final URI m_landingPageURI;
    private final URI m_loginURI;
    private final URI m_failureURI;
    private final String m_cookieName;
    private final String m_cookieDomain;
    private final String m_cookiePath;
    private final int m_cookieMaxAge;
    private final boolean m_endSessionAtLogout;
    private final boolean m_autoRenewToken;
    private final long m_tokenMaxLifetime;

    /**
     * Creates a new {@link AuthenticationResourceConfig} instance.
     */
    public AuthenticationResourceConfig() {
        m_landingPageURI = m_loginURI = m_failureURI = URI.create(DEFAULT_URI);
        m_cookieName = DEFAULT_TOKEN_COOKIE_NAME;
        m_cookieDomain = DEFAULT_COOKIE_DOMAIN;
        m_cookiePath = DEFAULT_COOKIE_PATH;
        m_cookieMaxAge = DEFAULT_COOKIE_MAX_AGE;
        m_endSessionAtLogout = DEFAULT_END_SESSION_AT_LOGOUT;
        m_autoRenewToken = DEFAULT_AUTO_RENEW_TOKEN;
        m_tokenMaxLifetime = DEFAULT_TOKEN_MAX_LIFETIME;
    }

    /**
     * Creates a new {@link AuthenticationResourceConfig} instance.
     */
    public AuthenticationResourceConfig(Dictionary<String, ?> config) throws ConfigurationException {
        m_landingPageURI = getURI(config, KEY_LANDING_PAGE_URL, DEFAULT_URI);
        m_loginURI = getURI(config, KEY_LOGIN_URL, DEFAULT_URI);
        m_failureURI = getURI(config, KEY_FAILURE_URL, DEFAULT_URI);
        m_cookieName = getString(config, KEY_COOKIE_NAME, DEFAULT_TOKEN_COOKIE_NAME);
        m_cookieDomain = getString(config, KEY_COOKIE_DOMAIN, DEFAULT_COOKIE_DOMAIN);
        m_cookiePath = getString(config, KEY_COOKIE_PATH, DEFAULT_COOKIE_PATH);
        m_cookieMaxAge = getInteger(config, KEY_COOKIE_MAX_AGE, DEFAULT_COOKIE_MAX_AGE);
        m_endSessionAtLogout = getBoolean(config, KEY_END_SESSION_AT_LOGOUT, DEFAULT_END_SESSION_AT_LOGOUT);
        m_autoRenewToken = getBoolean(config, KEY_AUTO_RENEW_TOKEN, DEFAULT_AUTO_RENEW_TOKEN);
        m_tokenMaxLifetime = getLong(config, KEY_TOKEN_MAX_LIFETIME, DEFAULT_TOKEN_MAX_LIFETIME);
    }

    /**
     * @return the domain of the cookie to return, can be <code>null</code>.
     */
    public String getCookieDomain() {
        return m_cookieDomain;
    }

    /**
     * @return the maximum age of returned cookies, in seconds. If &lt;= 0, the cookie will follow the lifetime of the identity token (if present).
     */
    public int getCookieMaxAge() {
        return m_cookieMaxAge;
    }

    /**
     * @return the name of the cookie to return, never <code>null</code>.
     */
    public String getCookieName() {
        return m_cookieName;
    }

    /**
     * @return the path of the cookie to return, never <code>null</code>.
     */
    public String getCookiePath() {
        return m_cookiePath;
    }

    /**
     * @return the URI to redirect to after an authentication failure.
     */
    public URI getFailureURL() {
        return m_failureURI;
    }

    /**
     * @return the URI to redirect to after a successful authentication.
     */
    public URI getLandingPageURL() {
        return m_landingPageURI;
    }

    /**
     * @return the URI to redirect to in case a login is necessary.
     */
    public URI getLoginURL() {
        return m_loginURI;
    }

    /**
     * @return the maximum lifetime allowed for a token that is auto renewed, -1 of no maximum is defined.
     */
    public long getTokenMaxLifetime() {
        return m_tokenMaxLifetime;
    }

    /**
     * @return <code>true</code> if expired token should be automatically be renewed in case the session still is active, <code>false</code> (the default) to automatically end the session when a token expires.
     */
    public boolean isAutoRenewToken() {
        return m_autoRenewToken;
    }

    /**
     * @return <code>true</code> if the "end-session" endpoint of the (remote) identity provider should be called upon logout, <code>false</code> if it should not be called.
     */
    public boolean isEndSessionAtLogout() {
        return m_endSessionAtLogout;
    }
}
