/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.authservice.rest.util;

import java.util.HashMap;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.core.NewCookie;

/**
 * Convenience methods for working with {@link HttpServletRequest}s.
 */
public class HttpUtil {
    public static final String CTX_REMOTE_ADDR = "remoteAddr";
    public static final String CTX_USER_AGENT = "userAgent";
    public static final String CTX_REQUESTED_WITH = "requestedWith";
    public static final String CTX_FORWARDED = "forwarded";

    private static final String USER_AGENT = "User-Agent";
    private static final String X_REQUESTED_WITH = "X-Requested-With";
    private static final String X_FORWARDED_PROTO = "X-Forwarded-Proto";

    private static final String XML_HTTP_REQUEST = "XMLHttpRequest";
    public static final String SET_COOKIE_HEADER = "Set-Cookie";

    /**
     * Determines and returns the context from the given servlet request.
     *
     * @param req the {@link HttpServletRequest} to extract the context from, cannot be <code>null</code>.
     * @return a map with the context from the given servlet request, never <code>null</code>.
     */
    public static Map<String, String> getContext(HttpServletRequest req) {
        Map<String, String> result = new HashMap<>();
        result.put(CTX_REMOTE_ADDR, req.getRemoteAddr());
        result.put(CTX_USER_AGENT, req.getHeader(USER_AGENT));
        result.put(CTX_FORWARDED, req.getHeader(X_FORWARDED_PROTO));
        result.put(CTX_REQUESTED_WITH, req.getHeader(X_REQUESTED_WITH));
        return result;
    }

    /**
     * Determines whether or not the given request is originating from an AJAX/XmlHttpRequest call.
     *
     * @param req the servlet request, cannot be <code>null</code>.
     * @return <code>true</code> if the given servlet request is an XML HTTP request, <code>false</code> otherwise.
     */
    public static boolean isXmlHttpRequest(HttpServletRequest req) {
        return XML_HTTP_REQUEST.equals(req.getHeader(X_REQUESTED_WITH));
    }

    public static String toHttpHeader(javax.ws.rs.core.Cookie jaxrsCookie) {
        StringBuilder sb = new StringBuilder();

        String name = jaxrsCookie.getName();
        if (name != null) {
            sb.append(name);
        }

        sb.append('=');

        String value = jaxrsCookie.getValue();
        if (value != null) {
            sb.append(value);
        }

        String domain = jaxrsCookie.getDomain();
        if (domain != null) {
            sb.append(";Domain=").append(domain);
        }

        String path = jaxrsCookie.getPath();
        if (path != null) {
            sb.append(";Path=").append(path);
        }

        if (jaxrsCookie instanceof javax.ws.rs.core.NewCookie) {
            NewCookie newCookie = (javax.ws.rs.core.NewCookie) jaxrsCookie;

            sb.append(";HttpOnly");

            boolean secure = newCookie.isSecure();
            if (secure) {
                sb.append(";Secure");
            }

            int maxAge = newCookie.getMaxAge();
            if (maxAge >= 0) {
                sb.append(";Max-Age=").append(maxAge);
            }
        }

        return sb.toString();
    }

}
