/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.authservice.session.file;

import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.toList;
import static org.osgi.service.log.LogService.LOG_WARNING;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.function.Supplier;

import org.amdatu.security.authentication.authservice.SessionStore;
import org.amdatu.security.authentication.authservice.SessionStoreException;
import org.apache.felix.dm.Component;
import org.osgi.framework.BundleContext;
import org.osgi.service.log.LogService;

/**
 * File based session store implementation.
 */
public class FileSessionStore implements SessionStore {
    private static final int FILE_MAGIC = 0x53455301;

    private final ReadWriteLock m_lock = new ReentrantReadWriteLock(false);
    private final Map<String, List<Session>> m_sessions = new HashMap<>();
    // Managed by Felix DM...
    private volatile BundleContext m_context;
    private volatile LogService m_log;

    @Override
    public boolean closeSession(String userId, String sessionId) {
        List<Session> newSessions = protectedRead(() -> {
            List<Session> curSessions = m_sessions.get(userId);
            if (curSessions == null) {
                return null;
            }

            List<Session> filteredSessions = curSessions.stream()
                .filter(s -> !s.getSessionId().equals(sessionId))
                .collect(toList());

            if (curSessions.size() == filteredSessions.size()) {
                // Session was not found, or already closed before...
                return null;
            }

            return filteredSessions;
        });

        if (newSessions == null) {
            return false;
        }

        return protectedWrite(() -> {
            m_sessions.put(userId, newSessions);
            return true;
        });
    }

    @Override
    public boolean containsSession(String userId, String sessionId) {
        return protectedRead(() -> {
            List<Session> sessions = m_sessions.getOrDefault(userId, emptyList());
            for (Session session : sessions) {
                if (session.getSessionId().equals(sessionId)) {
                    return true;
                }
            }
            return false;
        });
    }

    @Override
    public String createSession(String userId, Map<String, String> context) throws SessionStoreException {
        Session s = new Session(userId, generateSessionId(), context);

        return protectedWrite(() -> {
            if (!m_sessions.computeIfAbsent(userId, uid -> new ArrayList<>()).add(s)) {
                m_log.log(LOG_WARNING, "Failed to create session! Unable to add session to store!");
                throw new SessionStoreException(userId, "store_failed");
            }

            return s.getSessionId();
        });
    }

    protected final void clear() {
        m_sessions.clear();
    }

    protected void readAllSessions(File file) throws IOException {
        Map<String, List<Session>> newSessions = readSessions(file);
        m_sessions.clear();
        m_sessions.putAll(newSessions);
    }

    protected final int size() {
        return m_sessions.size();
    }

    /**
     * Called by Felix DM.
     */
    protected final void start(Component c) {
        File f = m_context.getDataFile("sessions.dat");
        if (f == null || !f.exists()) {
            return;
        }

        protectedWrite(() -> {
            try {
                readAllSessions(f);
            }
            catch (Exception e) {
                m_log.log(LOG_WARNING, "Failed to read all accounts!", e);
            }
            return null;
        });
    }

    /**
     * Called by Felix DM.
     */
    protected final void stop(Component c) {
        File f = m_context.getDataFile("accounts.dat");
        if (f == null || !f.canRead()) {
            return;
        }

        protectedWrite(() -> {
            try {
                writeAllSessions(f);
            }
            catch (Exception e) {
                m_log.log(LOG_WARNING, "Failed to write all accounts!", e);
            }
            return null;
        });
    }

    protected void writeAllSessions(File file) throws IOException {
        writeSessions(file, m_sessions);
    }

    private String generateSessionId() {
        return UUID.randomUUID().toString();
    }

    private <RT> RT protectedRead(Supplier<RT> supplier) {
        m_lock.readLock().lock();
        try {
            return supplier.get();
        }
        finally {
            m_lock.readLock().unlock();
        }
    }

    private <RT> RT protectedWrite(Supplier<RT> supplier) {
        m_lock.writeLock().lock();
        try {
            return supplier.get();
        }
        finally {
            m_lock.writeLock().unlock();
        }
    }

    private Map<String, List<Session>> readSessions(File file) throws IOException {
        try (FileInputStream fis = new FileInputStream(file);
                        DataInputStream dis = new DataInputStream(fis)) {
            int magic = dis.readInt();
            if (magic != FILE_MAGIC) {
                throw new IOException("Invalid accounts file: unexpected file magic!");
            }

            Map<String, List<Session>> result = new HashMap<>();

            int expected = dis.readInt();
            for (int i = 0; i < expected; i++) {
                Session session = new Session();
                session.read(dis);

                result.computeIfAbsent(session.getUserId(), uid -> new ArrayList<>()).add(session);
            }

            return result;
        }
    }

    private void writeSessions(File file, Map<String, List<Session>> sessions) throws IOException {
        try (FileOutputStream fos = new FileOutputStream(file);
                        DataOutputStream dos = new DataOutputStream(fos)) {
            dos.writeInt(FILE_MAGIC);

            // Calculate the total amount of sessions we'd to store...
            int total = sessions.values().stream()
                .mapToInt(List::size).sum();

            dos.writeInt(total);

            sessions.values().stream()
                .flatMap(ss -> ss.stream())
                .forEach(session -> {
                    try {
                        session.write(dos);
                    }
                    catch (IOException e) {
                        throw new RuntimeException("Failed to store session!", e);
                    }
                });
        }
    }
}
