/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.amdatu.security.authentication.authservice.session.file;

import static java.util.Collections.singletonMap;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.runners.MockitoJUnitRunner;
import org.osgi.framework.BundleContext;
import org.osgi.service.log.LogService;

/**
 * @author <a href="mailto:amdatu-developers@amdatu.org">Amdatu Project Team</a>
 */
@RunWith(MockitoJUnitRunner.class)
public class FileSessionStoreTest {

    @Mock
    private BundleContext m_context;
    @Mock
    private LogService m_logService;

    @InjectMocks
    private FileSessionStore m_backend = new FileSessionStore();

    @Rule
    public final ExpectedException m_rule = ExpectedException.none();

    @Test
    public void testReadWriteEmptySessionOk() throws Exception {
        File file = createTempFile();

        m_backend.writeAllSessions(file);

        m_backend.readAllSessions(file);

        assertEquals(0, m_backend.size());
    }

    @Test
    public void testReadWriteMultipleSessionOk() throws Exception {
        File file = createTempFile();

        Map<String, List<String>> sessions = new HashMap<>();
        for (int i = 0; i < 10; i++) {
            String userId = String.format("user%d", (i % 3));
            String sessionId = m_backend.createSession(userId, singletonMap("remoteAddr", "127.0.0.1"));

            sessions.computeIfAbsent(userId, uid -> new ArrayList<>()).add(sessionId);
        }

        m_backend.writeAllSessions(file);

        m_backend.clear();

        m_backend.readAllSessions(file);

        assertEquals(3, m_backend.size());

        for (Entry<String, List<String>> entry : sessions.entrySet()) {
            String userId = entry.getKey();
            for (String sessionId : entry.getValue())
                assertTrue(m_backend.containsSession(userId, sessionId));
        }
    }

    @Test
    public void testReadWriteSingleSessionOk() throws Exception {
        File file = createTempFile();

        String sessionId = m_backend.createSession("user", singletonMap("remoteAddr", "127.0.0.1"));

        m_backend.writeAllSessions(file);

        m_backend.clear();

        m_backend.readAllSessions(file);

        assertEquals(1, m_backend.size());
        assertTrue(m_backend.containsSession("user", sessionId));
    }

    private File createTempFile() throws IOException {
        File file = File.createTempFile("sessions", ".dat");
        file.deleteOnExit();
        return file;
    }
}
