/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.idprovider.local;

import static java.time.Instant.now;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_ACCESS_DENIED;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_CREDENTIALS_RESET;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_INVALID_REQUEST_URI;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_UNAUTHORIZED_CLIENT;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.ERROR_CODE;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.PROVIDER_TYPE;
import static org.amdatu.security.authentication.idprovider.local.LocalIdProviderConfig.TYPE_LOCAL;
import static org.amdatu.security.authentication.idprovider.local.util.ConfigUtils.getString;
import static org.amdatu.security.tokenprovider.TokenConstants.SUBJECT;
import static org.osgi.service.log.LogService.LOG_INFO;

import java.net.URI;
import java.util.Base64;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.amdatu.security.account.Account;
import org.amdatu.security.account.AccountAdmin;
import org.amdatu.security.account.AccountCredentialResetException;
import org.amdatu.security.account.AccountException;
import org.amdatu.security.authentication.idprovider.IdProvider;
import org.amdatu.security.authentication.idprovider.IdProviderConstants;
import org.amdatu.security.authentication.idprovider.UserLookupService;
import org.amdatu.web.util.URIBuilder;
import org.apache.felix.dm.Component;
import org.osgi.service.cm.ConfigurationException;
import org.osgi.service.cm.ManagedService;
import org.osgi.service.log.LogService;

/**
 * Provides an {@link IdProvider} implementation that is able to act as local identity provider.
 */
public class LocalIdProvider implements IdProvider, ManagedService {

    // Injected by Felix DM...
    private volatile UserLookupService m_userLookupService;
    private volatile AccountAdmin m_accountAdmin;
    private volatile LogService m_logService;
    // Locally managed...
    private volatile LocalIdProviderConfig m_config;

    private static String coerce(String[] input) {
        if (input.length < 1) {
            return "";
        }
        if (input.length == 1) {
            return input[0];
        }
        return String.join(",", input);
    }

    @Override
    public URI getAuthenticationURI(URI callbackURI, Map<String, String[]> params) {
        LocalIdProviderConfig cfg = m_config;

        URIBuilder authURIBuilder = new URIBuilder(callbackURI);

        try {
            Map<String, String> credentials = getCredentials(cfg, params);

            Account account = m_accountAdmin.getAccount(credentials)
                .orElseThrow(() -> new AccountException());

            // We'll give each caller its own set of secrets, allowing us to counter
            // CORS (Cross Origin Resource Sharing) attacks and minimise the risk of
            // stuff leaking accidentally...
            byte[] code = cfg.generateClientKey();

            // Use a time limited code that can be used to correlate calls between this
            // method and the #getUserIdentity method...
            authURIBuilder.appendToQuery("state", generateStateCode(cfg, account, code));
            authURIBuilder.appendToQuery("code", encodeCode(code));
        }
        catch (AccountCredentialResetException e) {
            authURIBuilder.appendToQuery("error", CODE_CREDENTIALS_RESET);
        }
        catch (AccountException e) {
            authURIBuilder.appendToQuery("error", CODE_UNAUTHORIZED_CLIENT);
        }

        return authURIBuilder.build();
    }

    @Override
    public Map<String, String> getUserIdentity(URI callbackURI, Map<String, String[]> params) {
        LocalIdProviderConfig cfg = m_config;

        String error = getString(params, "error");
        String state = getString(params, "state");
        String code = getString(params, "code");

        Map<String, String> tokenAttrs = new HashMap<>();
        tokenAttrs.put(PROVIDER_TYPE, TYPE_LOCAL);

        String accountId = null;

        if (state == null || code == null) {
            // These two parameters should always be present!
            error = CODE_INVALID_REQUEST_URI;
        }
        else {
            accountId = verifyStateCode(cfg, state, decodeCode(code));
        }

        if (error == null) {
            try {
                Account account = m_accountAdmin.getAccountById(accountId);

                Map<String, String> attrs = new HashMap<>(account.getCredentials());
                attrs.put(PROVIDER_TYPE, TYPE_LOCAL);

                m_userLookupService.getUserId(attrs)
                    .ifPresent(localUserId -> tokenAttrs.put(SUBJECT, localUserId));
                tokenAttrs.put(IdProviderConstants.ACCOUNT_ID, accountId);
            }
            catch (AccountCredentialResetException e) {
                error = CODE_CREDENTIALS_RESET;
            }
            catch (AccountException e) {
                error = CODE_UNAUTHORIZED_CLIENT;
            }
        }

        if (error != null) {
            tokenAttrs.put(ERROR_CODE, error);
        }
        else if (!tokenAttrs.containsKey(SUBJECT)) {
            tokenAttrs.put(ERROR_CODE, CODE_ACCESS_DENIED);
        }

        return tokenAttrs;
    }

    /**
     * Called by Felix DM.
     */
    @Override
    public final void updated(Dictionary<String, ?> config) throws ConfigurationException {
        LocalIdProviderConfig newConfig;
        if (config != null) {
            newConfig = new LocalIdProviderConfig(config);
        }
        else {
            newConfig = new LocalIdProviderConfig();
        }
        m_config = newConfig;
    }

    /**
     * Called by Felix DM.
     */
    protected final void start(Component comp) throws Exception {
        m_logService.log(LOG_INFO, "Local ID provider started");
    }

    /**
     * Called by Felix DM.
     */
    protected final void stop(Component comp) throws Exception {
        m_logService.log(LOG_INFO, "Local ID provider stopped");
    }

    private byte[] decodeCode(String code) {
        return Base64.getUrlDecoder().decode(code);
    }

    private String encodeCode(byte[] code) {
        return Base64.getUrlEncoder().withoutPadding().encodeToString(code);
    }

    private String generateStateCode(LocalIdProviderConfig cfg, Account account, byte[] code) {
        return cfg.generateStateCodeAt(now(), code, account.getId());
    }

    private Map<String, String> getCredentials(LocalIdProviderConfig cfg, Map<String, String[]> params) {
        Set<String> wantedCredentials = cfg.getCredentialKeys();

        Map<String, String> credentials = new HashMap<>();

        params.entrySet().stream()
            .filter(e -> wantedCredentials.contains(e.getKey()))
            .forEach(e -> credentials.put(e.getKey(), coerce(e.getValue())));

        // Make sure the consumers of these credentials can determine where they came from...
        credentials.put(PROVIDER_TYPE, TYPE_LOCAL);

        return credentials;
    }

    private String verifyStateCode(LocalIdProviderConfig cfg, String state, byte[] code) {
        return cfg.verifyStateCodeAt(now(), code, state);
    }
}
