/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.idprovider.local;

import static java.util.Arrays.asList;
import static org.amdatu.security.authentication.idprovider.local.util.ConfigUtils.getInteger;
import static org.amdatu.security.authentication.idprovider.local.util.ConfigUtils.getMandatoryStringArray;
import static org.amdatu.security.authentication.idprovider.local.util.ConfigUtils.getMandatoryURI;
import static org.amdatu.security.authentication.idprovider.local.util.ConfigUtils.getString;
import static org.amdatu.security.util.crypto.CryptoUtils.createSecureRandom;
import static org.amdatu.security.util.crypto.CryptoUtils.generateSecret;
import static org.amdatu.security.util.crypto.CryptoUtils.getSecretKey;
import static org.jose4j.jwe.ContentEncryptionAlgorithmIdentifiers.AES_128_CBC_HMAC_SHA_256;
import static org.jose4j.jwe.KeyManagementAlgorithmIdentifiers.DIRECT;

import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.time.Instant;
import java.util.Dictionary;
import java.util.HashSet;
import java.util.Set;

import org.amdatu.security.password.util.Hkdf;
import org.amdatu.security.password.util.Totp;
import org.jose4j.jwe.JsonWebEncryption;
import org.jose4j.keys.AesKey;
import org.jose4j.lang.JoseException;
import org.osgi.service.cm.ConfigurationException;

/**
 * Represents the configuration for {@link LocalIdProvider}.
 */
public class LocalIdProviderConfig {
    /**
     * REQUIRED. An array of key-names to expect as credentials that should be used to authenticate a user.
     */
    public static final String KEY_CREDENTIAL_KEYS = "credentialKeys";
    /**
     * REQUIRED. The URI containing or referencing the secret key used for encrypting the state value of each login.
     */
    public static final String KEY_KEY_URI = "keyUri";
    /**
     * OPTIONAL. The time window for which a generated authenticator token is valid, in seconds. Defaults to 10 seconds.
     */
    public static final String KEY_TIMESTEP_WINDOW = "timestepWindow";
    /**
     * OPTIONAL. The algorithm used for the {@link SecureRandom} instance. Defaults to using the default algorithm selected by {@code new SecureRandom()}.
     */
    public static final String KEY_PRNG_ALGORITHM = "secureRandomAlgorithm";

    public static final String TYPE_LOCAL = "local";

    public static final String HMAC_SHA256 = "HmacSHA256";
    public static final int SALT_SECRET_LENGTH = 8;
    public static final int SECRET_KEY_LENGTH = 32;
    public static final int TIMESTEP_WINDOW = 10;
    public static final int CLIENT_KEY_LENGTH = 16;

    private final Set<String> m_credentialKeys;
    private final byte[] m_secretKey;
    private final SecureRandom m_prng;
    private final int m_timestepWindow;

    /**
     * Creates a new {@link LocalIdProviderConfig} instance.
     */
    public LocalIdProviderConfig() {
        m_prng = new SecureRandom();
        m_credentialKeys = new HashSet<>(asList("email", "password"));
        m_secretKey = generateSecret(m_prng, 64);
        m_timestepWindow = TIMESTEP_WINDOW;
    }

    public LocalIdProviderConfig(Dictionary<String, ?> config) throws ConfigurationException {
        m_credentialKeys = new HashSet<>(asList(getMandatoryStringArray(config, KEY_CREDENTIAL_KEYS)));
        m_timestepWindow = getInteger(config, KEY_TIMESTEP_WINDOW, TIMESTEP_WINDOW);

        try {
            m_prng = createSecureRandom(getString(config, KEY_PRNG_ALGORITHM, null));

            m_secretKey = getSecretKey(m_prng, getMandatoryURI(config, KEY_KEY_URI));
        }
        catch (NoSuchAlgorithmException e) {
            throw new ConfigurationException(null, "unsupported algorithm for PRNG or secret key!", e);
        }
        catch (IOException e) {
            throw new ConfigurationException(KEY_KEY_URI, "invalid file or resource!", e);
        }
    }

    /**
     * @return a random client key, never <code>null</code>.
     */
    public byte[] generateClientKey() {
        return generateSecret(m_prng, CLIENT_KEY_LENGTH);
    }

    /**
     * Generates a state-value that captures the given payload.
     * <p>
     * This method encrypts the given payload using a derived key and returns this value as JWE
     * token using compact serialization.
     * </p>
     *
     * @param timestamp the timestamp to use in the key derivation process;
     * @param clientKey the code generated for each authentication request;
     * @param payload the payload to embed in the generated state value.
     * @return the generated state value.
     */
    public String generateStateCodeAt(Instant timestamp, byte[] clientKey, String payload) {
        try {
            JsonWebEncryption jwe = new JsonWebEncryption();
            jwe.setKey(new AesKey(deriveSecret(timestamp, clientKey)));
            jwe.setEncryptionMethodHeaderParameter(AES_128_CBC_HMAC_SHA_256);
            jwe.setAlgorithmHeaderValue(DIRECT);
            jwe.setPayload(payload);

            return jwe.getCompactSerialization();
        }
        catch (JoseException e) {
            // We only get this exception in case our derived key is not of the expected length!
            throw new RuntimeException("Invalid configuration for generating state values!", e);
        }
    }

    /**
     * @return the credential keys to expect when authenticating a user.
     */
    public Set<String> getCredentialKeys() {
        return m_credentialKeys;
    }

    /**
     * @param clientKey the code generated for each authentication request;
     * @param state the actual state value to verify.
     * @return the payload of the given state value.
     */
    public String verifyStateCodeAt(Instant timestamp, byte[] clientKey, String state) {
        try {
            JsonWebEncryption jwe = new JsonWebEncryption();
            jwe.setKey(new AesKey(deriveSecret(timestamp, clientKey)));
            jwe.setCompactSerialization(state);

            return jwe.getPayload();
        }
        catch (JoseException e) {
            // The contents were bogus or too old, do not return anything...
            return null;
        }
    }

    /**
     * Derives a secret key for the given client key and the secret key of this instance.
     * <p>
     * This method works in two steps. First, it creates a time-based token that is used as salting
     * for actual key derivation step. If, within a certain timestep (defaults to 10 seconds), this
     * method is called again, the same salt will be generated for the same input parameters. This
     * way, we do not need to return the actual salt to the client. In addition, assuming the given
     * timestamp increases, the salt will result in different derived keys over time.<br>
     * Once the salt is generated, it is used to derive a key based on the shared secret key of
     * this instance, and the given client key. This ensures that unique keys are generated for
     * each unique client key without leaking actual information about those keys.
     * </p>
     *
     * @param timestamp the timestamp to use in the derivation process;
     * @param clientKey the client key to use, should not be reused for multiple authentication attempts.
     * @return the secret key material to use for encrypting the state value.
     */
    protected byte[] deriveSecret(Instant timestamp, byte[] clientKey) {
        Totp otp = new Totp.Builder()
            .setAlgorithm(HMAC_SHA256)
            .deriveSecret(m_secretKey, clientKey, SALT_SECRET_LENGTH)
            .setTimeStep(m_timestepWindow)
            .build();

        byte[] salt = otp.generateRawCode(timestamp.getEpochSecond(), new byte[0], null);

        Hkdf kdf = new Hkdf.Builder()
            .setAlgorithm(HMAC_SHA256)
            .setInitialKeyingMaterial(m_secretKey)
            .setSalt(salt)
            .build();

        return kdf.deriveKey(clientKey, SECRET_KEY_LENGTH);
    }

}
