/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.idprovider.openid;

import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_ACCESS_DENIED;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.ERROR_CODE;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.PROVIDER_TYPE;
import static org.amdatu.security.authentication.idprovider.openid.util.ConfigUtil.updateServiceProperties;
import static org.amdatu.security.tokenprovider.TokenConstants.SUBJECT;
import static org.osgi.service.log.LogService.LOG_INFO;

import java.net.URI;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.Map;

import org.amdatu.security.authentication.idprovider.IdProvider;
import org.amdatu.security.authentication.idprovider.UserLookupService;
import org.apache.felix.dm.Component;
import org.openid4java.association.AssociationException;
import org.openid4java.consumer.ConsumerException;
import org.openid4java.consumer.VerificationResult;
import org.openid4java.discovery.DiscoveryException;
import org.openid4java.discovery.Identifier;
import org.openid4java.message.MessageException;
import org.osgi.service.log.LogService;

/**
 * Provides an OpenID identity provider implementation.
 */
public class OpenIdProvider implements IdProvider {

    // Injected by Felix DM...
    private volatile UserLookupService m_userLookupService;
    private volatile LogService m_log;
    private volatile Component m_component;
    // Locally managed...
    private volatile OpenIdProviderConfig m_config;

    /**
     * Creates a new {@link OpenIdProvider} instance.
     */
    public OpenIdProvider() {
        // Nop
    }

    @Override
    public URI getAuthenticationURI(URI callbackURI, Map<String, String[]> params) {
        try {
            return URI.create(m_config.createAuthURI(callbackURI));
        }
        catch (MessageException | ConsumerException | DiscoveryException e) {
            m_log.log(LOG_INFO, "Failed to get authentication URI!", e);
            throw new RuntimeException();
        }
    }

    @Override
    public Map<String, String> getUserIdentity(URI callbackURI, Map<String, String[]> params) {
        OpenIdProviderConfig cfg = m_config;

        Map<String, String> tokenAttrs = new HashMap<>();
        tokenAttrs.put(PROVIDER_TYPE, cfg.getType());

        String error = null;

        try {
            VerificationResult verificationResult = cfg.verifyResult(callbackURI, params);

            Identifier verifiedId = verificationResult.getVerifiedId();

            error = verificationResult.getStatusMsg();

            Map<String, String> attrs = new HashMap<>();
            attrs.put(PROVIDER_TYPE, cfg.getType());
            attrs.put("verifiedId", verifiedId.getIdentifier());

            m_userLookupService.getUserId(attrs)
                .ifPresent(userId -> tokenAttrs.put(SUBJECT, userId));
        }
        catch (DiscoveryException | MessageException | AssociationException e) {
            m_log.log(LOG_INFO, "Failed to get user identity!", e);
            throw new RuntimeException();
        }

        if (error != null) {
            tokenAttrs.put(ERROR_CODE, error);
        }
        else if (!tokenAttrs.containsKey(SUBJECT)) {
            tokenAttrs.put(ERROR_CODE, CODE_ACCESS_DENIED);
        }

        return tokenAttrs;
    }

    /**
     * Called by Felix DM.
     */
    protected final void start(Component comp) throws Exception {
        m_log.log(LOG_INFO, "OpenId ID provider started for type: " + m_config.getType());
    }

    /**
     * Called by Felix DM.
     */
    protected final void stop(Component comp) throws Exception {
        m_log.log(LOG_INFO, "OpenId ID provider stopped for type: " + m_config.getType());
    }

    /**
     * Called by Felix DM.
     */
    protected final void update(Dictionary<String, ?> config) throws Exception {
        OpenIdProviderConfig newConfig = null;
        if (config != null) {
            newConfig = new OpenIdProviderConfig(config);
        }
        m_config = newConfig;

        if (newConfig != null) {
            updateServiceProperties(m_component, PROVIDER_TYPE, newConfig.getType());
        }
    }
}
