/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.idprovider.oidc;

import static org.amdatu.security.authentication.idprovider.IdProviderConstants.ACCESS_TOKEN;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_ACCESS_DENIED;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_INVALID_REQUEST_URI;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.ERROR_CODE;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.ID_TOKEN;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.PROMPT;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.PROVIDER_TYPE;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.REFRESH_TOKEN;
import static org.amdatu.security.authentication.idprovider.oidc.util.ConfigUtil.getString;
import static org.amdatu.security.authentication.idprovider.oidc.util.ConfigUtil.updateServiceProperties;
import static org.amdatu.security.tokenprovider.TokenConstants.EXPIRATION_TIME;
import static org.amdatu.security.tokenprovider.TokenConstants.SUBJECT;
import static org.osgi.service.log.LogService.LOG_INFO;
import static org.osgi.service.log.LogService.LOG_WARNING;

import java.net.URI;
import java.time.Instant;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.amdatu.security.authentication.idprovider.IdProvider;
import org.amdatu.security.authentication.idprovider.UserLookupService;
import org.amdatu.security.tokenprovider.InvalidTokenException;
import org.amdatu.security.tokenprovider.TokenProvider;
import org.amdatu.web.util.URIBuilder;
import org.apache.felix.dm.Component;
import org.osgi.service.log.LogService;

import com.github.scribejava.core.model.OAuth2AccessToken;
import com.github.scribejava.core.oauth.OAuth20Service;

/**
 * Provides a generic OpenID connect-based {@link IdProvider} implementation.
 */
public class OpenIdConnectProvider implements IdProvider {
    // Injected by Felix DM...
    private volatile UserLookupService m_userLookupService;
    private volatile TokenProvider m_tokenProvider;
    private volatile LogService m_log;
    private volatile Component m_component;
    // Locally managed...
    private volatile OpenIdConnectProviderConfig m_config;

    /**
     * Creates a new {@link OpenIdConnectProvider} instance.
     */
    public OpenIdConnectProvider() {
        // Nop
    }

    @Override
    public URI getAuthenticationURI(URI callbackURI, Map<String, String[]> params) {
        OpenIdConnectProviderConfig cfg = m_config;

        // We'll give each caller its own secret, allowing us to counter XSRF attacks...
        String state = cfg.generateStateCode();

        OAuth20Service oauthSrv = cfg.createOAuthServiceWithState(callbackURI, state);

        Map<String, String> oauthParams = cfg.getAdditionalOAuthParams();
        // If there's a prompt specified in the given parameters, use it as part of the
        // OAuth/OpenID connect authentication URL. It allows us to control the login screen
        // presented to the user when calling the remote OpenID connect provider...
        String[] promptParam = params.getOrDefault(PROMPT, new String[0]);
        if (promptParam.length > 0) {
            oauthParams.put(PROMPT, String.join(" ", promptParam));
        }

        return URI.create(oauthSrv.getAuthorizationUrl(oauthParams));
    }

    @Override
    public Optional<URI> getEndSessionURI(URI callback, Map<String, String> credentials) {
        String accessToken = credentials.get(ACCESS_TOKEN);
        if (accessToken == null || "".equals(accessToken)) {
            return Optional.empty();
        }

        return m_config.getSessionEndURI()
            .map(uriBuilder -> decorateURI(uriBuilder, callback, credentials));
    }

    @Override
    public Map<String, String> getUserIdentity(URI callbackURI, Map<String, String[]> params) {
        OpenIdConnectProviderConfig cfg = m_config;

        String error = getString(params, "error");
        String state = getString(params, "state");
        String code = getString(params, "code");

        Map<String, String> tokenAttrs = new HashMap<>();
        tokenAttrs.put(PROVIDER_TYPE, cfg.getType());

        // Provide an additional check (XSRF)...
        if (error == null && (state == null || !cfg.verifyStateCode(state))) {
            m_log.log(LOG_WARNING, String.format("Unable to verify callback request for %s from %s: state mismatch! "
                + "Request took either too long or possible replay attack?!", params, cfg.getType()));

            error = CODE_INVALID_REQUEST_URI;
        }

        OAuth20Service oauthService = cfg.createOAuthServiceWithState(callbackURI, state);
        if (error == null) {
            OAuth2AccessToken token = oauthService.getAccessToken(code);

            tokenAttrs.put(ACCESS_TOKEN, token.getAccessToken());

            String refreshToken = token.getRefreshToken();
            if (refreshToken != null) {
                tokenAttrs.put(REFRESH_TOKEN, refreshToken);
            }

            String idToken = token.getParameter(ID_TOKEN);
            if (idToken != null) {
                Map<String, String> attrs = m_tokenProvider.verifyToken(idToken);
                // Give the user lookup service a clue where these attributes came from...
                attrs.put(PROVIDER_TYPE, cfg.getType());

                List<String> wantedClaims = cfg.getClaims();

                m_userLookupService.getUserId(attrs)
                    .ifPresent(userId -> {
                        attrs.entrySet().stream()
                            .filter(e -> wantedClaims.contains(e.getKey()))
                            .forEach(e -> tokenAttrs.put(e.getKey(), e.getValue()));

                        tokenAttrs.put(SUBJECT, userId);
                    });
            }
        }

        if (error != null) {
            tokenAttrs.put(ERROR_CODE, error);
        }
        else if (!tokenAttrs.containsKey(SUBJECT)) {
            tokenAttrs.put(ERROR_CODE, CODE_ACCESS_DENIED);
        }

        return tokenAttrs;
    }

    @Override
    public boolean refreshToken(Map<String, String> credentials) throws InvalidTokenException {
        String refreshToken = credentials.get(REFRESH_TOKEN);
        if (refreshToken == null || "".equals(refreshToken.trim())) {
            return false;
        }

        String exp = credentials.get(EXPIRATION_TIME);
        if (exp != null) {
            Instant inst = Instant.ofEpochSecond(Long.valueOf(exp));
            if (Instant.now().minusSeconds(60).isBefore(inst)) {
                return false;
            }
        }

        OAuth20Service oauthSrv = m_config.createOAuthService();
        OAuth2AccessToken token = oauthSrv.refreshAccessToken(refreshToken);
        if (token != null) {
            credentials.put(ACCESS_TOKEN, token.getAccessToken());
            return true;
        }

        return false;
    }

    /**
     * Called by Felix DM.
     */
    protected final void start(Component comp) throws Exception {
        m_log.log(LOG_INFO, "OpenId connect ID provider started for type: " + m_config.getType());
    }

    /**
     * Called by Felix DM.
     */
    protected final void stop(Component comp) throws Exception {
        m_log.log(LOG_INFO, "OpenId connect ID provider stopped for type: " + m_config.getType());
    }

    /**
     * Called by Felix DM.
     */
    protected final void update(Dictionary<String, ?> config) throws Exception {
        OpenIdConnectProviderConfig newConfig = null;
        if (config != null) {
            newConfig = new OpenIdConnectProviderConfig(config);
        }
        m_config = newConfig;

        if (newConfig != null) {
            updateServiceProperties(m_component,
                PROVIDER_TYPE, newConfig.getType(),
                "issuer", newConfig.getIdTokenIssuer(),
                "url", newConfig.getIdTokenJwksUri());
        }
    }

    private URI decorateURI(URIBuilder builder, URI callback, Map<String, String> credentials) {
        if (callback != null) {
            builder.appendToQuery("post_logout_redirect_uri", callback.toASCIIString());
        }
        return builder.build();
    }
}
