/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.idprovider.oidc;

import static java.util.Collections.list;
import static org.amdatu.security.authentication.idprovider.oidc.util.ConfigUtil.getMandatoryString;
import static org.amdatu.security.authentication.idprovider.oidc.util.ConfigUtil.getString;
import static org.amdatu.security.authentication.idprovider.oidc.util.ConfigUtil.getStringArray;
import static org.amdatu.security.authentication.idprovider.oidc.util.ConfigUtil.getURI;
import static org.amdatu.security.util.crypto.CryptoUtils.createSecureRandom;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URL;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.Arrays;
import java.util.Base64;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.amdatu.security.password.util.Totp;
import org.amdatu.web.util.URIBuilder;
import org.osgi.service.cm.ConfigurationException;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.PropertyNamingStrategy;
import com.github.scribejava.apis.OpenIdConnectApi;
import com.github.scribejava.core.builder.ServiceBuilder;
import com.github.scribejava.core.model.OpenIdConnectConfig;
import com.github.scribejava.core.oauth.OAuth20Service;

/**
 * Represents the {@link OpenIdConnectProvider} configuration.
 */
public class OpenIdConnectProviderConfig {
    /**
     * REQUIRED. The remote ID provider type, like "google", that can be used to reference this ID provider.
     */
    public static final String KEY_PROVIDER_TYPE = "providerType";
    /**
     * REQUIRED. The client or API identifier for the remote ID provider.
     */
    public static final String KEY_CLIENT_ID = "clientId";
    /**
     * REQUIRED. The client or API secret for the remote ID provider.
     */
    public static final String KEY_CLIENT_SECRET = "clientSecret";
    /**
     * OPTIONAL. The claims to copy from the ID token, if this one is returned.
     */
    public static final String KEY_CLAIMS = "claims";
    /**
     * OPTIONAL. The type of grant to request from the remote ID provider. Defaults to <tt>authorization_code</tt>.
     */
    public static final String KEY_GRANT_TYPE = "grantType";
    /**
     * OPTIONAL. The JSON-string describing the OpenID connect discovery information. If omitted uses either the URL provided by {@link #KEY_OPEN_ID_CONFIG_URL} or determines it based on the given {@link #KEY_PROVIDER_TYPE}.
     */
    public static final String KEY_OPEN_ID_CONFIG = "openIdConfig";
    /**
     * OPTIONAL. The URL to the OpenID connect discovery endpoint. If omitted uses either the {@link #KEY_OPEN_ID_CONFIG} or determines the endpoint based on the {@link #KEY_PROVIDER_TYPE}.
     */
    public static final String KEY_OPEN_ID_CONFIG_URL = "openIdConfigURL";
    /**
     * OPTIONAL. The type of response to request from the remote ID provider. Defaults to <tt>code</tt>.
     */
    public static final String KEY_RESPONSE_TYPE = "responseType";
    /**
     * OPTIONAL. The scopes to request access to. Multiple scopes should be separated by single spaces. Defaults to <tt>openid email profile</tt>.
     */
    public static final String KEY_SCOPES = "scopes";
    /**
     * OPTIONAL. The algorithm used for the {@link SecureRandom} instance. Defaults to using the default algorithm selected by {@code new SecureRandom()}.
     */
    public static final String KEY_PRNG_ALGORITHM = "secureRandomAlgorithm";

    /**
     * Prefix used to pass additional parameters to the OAuth authorization URL.
     */
    private static final String KEY_EXTRA_PARAM_PREFIX = "param.";

    public static final String TYPE_GOOGLE = "google";
    public static final String TYPE_OFFICE = "office";

    private static final String DEFAULT_SCOPES = "openid email profile";
    private static final String DEFAULT_GRANT_TYPE = "authorization_code";
    private static final String DEFAULT_RESPONSE_TYPE = "code";
    private static final String[] DEFAULT_STD_CLAIMS = {
        "sub", "name", "given_name", "family_name", "middle_name", "nickname", "preferred_username", "profile",
        "picture", "website", "email", "email_verified", "gender", "birthdate", "zoneinfo", "locale",
        "phone_number", "phone_number_verified", "address", "updated_at"
    };

    private static final int SALT_LENGTH = 8;

    /** OpenID connect discovery URL for Google. */
    static final String GOOGLE_OPENID_CONF = "https://accounts.google.com/.well-known/openid-configuration";
    /** OpenID connect discovery URL for Office365. */
    static final String OFFICE_OPENID_CONF = "https://login.microsoftonline.com/common/.well-known/openid-configuration";

    private final String m_type;

    private final String m_apiKey;
    private final String m_apiSecret;
    private final String m_grantType;
    private final String m_responseType;
    private final String m_scopes;
    private final Map<String, String> m_extraAuthParams;

    private final SecureRandom m_prng;

    private final List<String> m_claims;
    private final OpenIdConnectConfig m_apiConfig;
    private final Totp m_totp;

    /**
     * Creates a new {@link OpenIdConnectProviderConfig} instance.
     */
    public OpenIdConnectProviderConfig(Dictionary<String, ?> config) throws ConfigurationException, IOException {
        m_type = getMandatoryString(config, KEY_PROVIDER_TYPE);

         try {
            m_prng = createSecureRandom(getString(config, KEY_PRNG_ALGORITHM, null));
        }
        catch (NoSuchAlgorithmException e) {
            throw new ConfigurationException(KEY_PRNG_ALGORITHM, "unsupported algorithm!", e);
        }

        m_apiKey = getMandatoryString(config, KEY_CLIENT_ID);
        m_apiSecret = getMandatoryString(config, KEY_CLIENT_SECRET);
        m_grantType = getString(config, KEY_GRANT_TYPE, DEFAULT_GRANT_TYPE);
        m_responseType = getString(config, KEY_RESPONSE_TYPE, DEFAULT_RESPONSE_TYPE);
        m_scopes = getString(config, KEY_SCOPES, DEFAULT_SCOPES);

        m_claims = Arrays.asList(getStringArray(config, KEY_CLAIMS, DEFAULT_STD_CLAIMS));

        m_extraAuthParams = new HashMap<>();
        for (String key : list(config.keys())) {
            if (!key.startsWith(KEY_EXTRA_PARAM_PREFIX)) {
                continue;
            }
            String val = getString(config, key, null);
            if (val != null) {
                m_extraAuthParams.put(key.substring(KEY_EXTRA_PARAM_PREFIX.length()), val);
            }
        }

        m_apiConfig = getOpenIdConnectConfig(config);

        m_totp = new Totp.Builder().deriveSecret(m_apiSecret, m_apiKey, 16).build();
    }

    static OpenIdConnectConfig getOpenIdConnectConfig(Dictionary<String, ?> dict) throws ConfigurationException, IOException {
        String config = getString(dict, KEY_OPEN_ID_CONFIG, null);
        if (config != null) {
            return parseOICConfig(new ByteArrayInputStream(config.getBytes()));
        }
        URI configURI = getURI(dict, KEY_OPEN_ID_CONFIG_URL, null);
        if (configURI == null) {
            String type = getMandatoryString(dict, KEY_PROVIDER_TYPE).toLowerCase();
            if (TYPE_GOOGLE.equals(type)) {
                configURI = URI.create(GOOGLE_OPENID_CONF);
            }
            else if (TYPE_OFFICE.equals(type)) {
                configURI = URI.create(OFFICE_OPENID_CONF);
            }
        }
        if (configURI == null) {
            throw new ConfigurationException(null, "neither '" + KEY_OPEN_ID_CONFIG + "' nor '" + KEY_OPEN_ID_CONFIG_URL + "' is supplied!");
        }

        return parseOICConfig(configURI);
    }

    static OpenIdConnectConfig parseOICConfig(InputStream is) throws IOException {
        // The OpenIdConnectConfig uses camel casing, but the JSON format used in the OpenID connect
        // discovery uses snake-casing (with underscores). We need to convert between the two.
        // In addition, the OpenID connect discovery information can contain additional properties
        // which we should ignore.
        ObjectMapper mapper = new ObjectMapper()
            .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false)
            .setPropertyNamingStrategy(PropertyNamingStrategy.CAMEL_CASE_TO_LOWER_CASE_WITH_UNDERSCORES);
        return mapper.readValue(is, OpenIdConnectConfig.class);
    }

    static OpenIdConnectConfig parseOICConfig(URI configURI) throws IOException {
        URL url = configURI.toURL();
        HttpURLConnection conn = (HttpURLConnection) url.openConnection();
        try {
            return parseOICConfig(conn.getInputStream());
        }
        finally {
            conn.disconnect();
        }
    }

    public OAuth20Service createOAuthService() {
        return createOAuthServiceWithState(null, null);
    }

    public OAuth20Service createOAuthServiceWithState(URI callbackURI, String state) {
        ServiceBuilder serviceBuilder = new ServiceBuilder();
        serviceBuilder.apiKey(m_apiKey);
        serviceBuilder.apiSecret(m_apiSecret);
        serviceBuilder.grantType(m_grantType);
        serviceBuilder.responseType(m_responseType);
        serviceBuilder.scope(m_scopes);
        if (callbackURI != null) {
            serviceBuilder.callback(callbackURI.toASCIIString());
        }
        if (state != null) {
            serviceBuilder.state(state);
        }
        return serviceBuilder.build(new OpenIdConnectApi(m_apiConfig));
    }

    public String generateStateCode() {
        byte[] salt = new byte[SALT_LENGTH];
        m_prng.nextBytes(salt);

        byte[] dest = new byte[salt.length + m_totp.getMacLength()];

        m_totp.generateRawCode(salt, dest);
        // append the salt for our own use later on...
        System.arraycopy(salt, 0, dest, dest.length - salt.length, salt.length);

        return Base64.getUrlEncoder().withoutPadding().encodeToString(dest);
    }

    public boolean verifyStateCode(String code) {
        byte[] data = Base64.getUrlDecoder().decode(code);
        if (data.length < SALT_LENGTH) {
            return false;
        }

        byte[] salt = Arrays.copyOfRange(data, data.length - SALT_LENGTH, data.length);
        byte[] rawCode = Arrays.copyOf(data, data.length - SALT_LENGTH);

        return m_totp.validateRawCode(rawCode, salt);
    }

    public Map<String, String> getAdditionalOAuthParams() {
        return m_extraAuthParams;
    }

    /**
     * @return the claims to extract from the ID token, never <code>null</code>.
     */
    public List<String> getClaims() {
        return m_claims;
    }

    /**
     * @return the issuer of the ID tokens returned by the OpenID connect provider.
     */
    public String getIdTokenIssuer() {
        String issuer = m_apiConfig.getIssuer();
        Pattern p = Pattern.compile("\\{[^}]+\\}");
        Matcher m = p.matcher(issuer);
        return m.replaceAll(".*?");
    }

    /**
     * @return the URI to the JSON web keys that can be used to validate the ID tokens, never <code>null</code>.
     */
    public String getIdTokenJwksUri() {
        return m_apiConfig.getJwksUri();
    }

    /**
     * @return an optional URL to use for logging out the user.
     */
    public Optional<URIBuilder> getSessionEndURI() {
        String url = m_apiConfig.getEndSessionEndpoint();
        if (url == null || "".equals(url)) {
            return Optional.empty();
        }
        return Optional.of(new URIBuilder(url));
    }

    /**
     * @return the type of OpenID connect provider, used to find back a particular provider.
     */
    public String getType() {
        return m_type;
    }
}
