/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.test.app;

import static java.util.Arrays.asList;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CountDownLatch;

import org.amdatu.security.account.Account;
import org.amdatu.security.account.AccountAdminBackend;
import org.amdatu.security.account.AccountExistsException;
import org.amdatu.security.account.Account.State;
import org.amdatu.security.authentication.authservice.AuthenticationService;
import org.amdatu.security.authentication.idprovider.UserLookupService;

/**
 * Mock implementation of {@link AuthenticationService}.
 */
public class MyAuthService implements AuthenticationService, UserLookupService, AccountAdminBackend {
    public final CountDownLatch m_loginCalled = new CountDownLatch(1);
    public final CountDownLatch m_logoutCalled = new CountDownLatch(1);
    public final CountDownLatch m_getUserIdCalled = new CountDownLatch(1);

    private final Set<String> m_credentialKeys;
    private final Set<String> m_knownUserIds;
    private final Map<String, Account> m_accounts;

    /**
     * Creates a new {@link MyAuthService} instance.
     */
    public MyAuthService(String... knownUserIds) {
        m_credentialKeys = new HashSet<>(asList("email", "upn"));
        m_knownUserIds = new HashSet<>(asList(knownUserIds));
        m_accounts = new HashMap<>();
        Map<String, String> creds = new HashMap<>();
        creds.put("email", "known_user@localhost");
        creds.put("password", "secret");
        m_accounts.put("known_user@localhost", new Account("known_user@localhost", creds, State.NORMAL, null, false));
    }

    @Override
    public boolean accountExists(String accountId) {
        return m_accounts.containsKey(accountId);
    }

    @Override
    public void create(Account account) throws AccountExistsException {
        if (m_accounts.containsKey(account.getId())) {
            throw new AccountExistsException();
        }
        m_accounts.putIfAbsent(account.getId(), account);
    }

    @Override
    public Optional<Account> get(String accountId) {
        return Optional.ofNullable(m_accounts.get(accountId));
    }

    @Override
    public Optional<String> getUserId(Map<String, String> credentials) {
        m_getUserIdCalled.countDown();

        for (String credentialKey : m_credentialKeys) {
            String userName = credentials.get(credentialKey);
            if (m_knownUserIds.contains(userName)) {
                return Optional.of(userName);
            }
        }

        return Optional.empty();
    }

    @Override
    public boolean isLoggedIn(Map<String, String> credentials) {
        return m_loginCalled.getCount() == 0L;
    }

    @Override
    public void login(Map<String, String> credentials, Map<String, String> context) {
        m_loginCalled.countDown();
        System.out.println("LOGIN FOR " + credentials);
    }

    @Override
    public void logout(Map<String, String> credentials) {
        m_logoutCalled.countDown();
        System.out.println("LOGOUT FOR " + credentials);
    }

    @Override
    public void update(Account account) {
        m_accounts.put(account.getId(), account);
    }

}
