/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.test.authservice;

import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_INVALID_REQUEST;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.CODE_LOGIN_REQUIRED;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.PROVIDER_TYPE;
import static org.amdatu.security.authentication.test.authservice.TestAuthService.KNOWN_USER;
import static org.amdatu.security.authentication.test.authservice.TestAuthService.SESSION_ID;
import static org.amdatu.security.tokenprovider.TokenConstants.EXPIRATION_TIME;
import static org.amdatu.security.tokenprovider.TokenConstants.NOT_BEFORE;
import static org.amdatu.security.tokenprovider.TokenConstants.SUBJECT;
import static org.amdatu.testing.configurator.TestConfigurator.cleanUp;
import static org.amdatu.testing.configurator.TestConfigurator.configure;
import static org.amdatu.testing.configurator.TestConfigurator.createComponent;
import static org.amdatu.testing.configurator.TestConfigurator.createServiceDependency;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Matchers.anyMapOf;
import static org.mockito.Matchers.anyObject;
import static org.mockito.Matchers.anyString;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

import java.io.IOException;
import java.time.Instant;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.amdatu.security.authentication.authservice.AuthenticationHandler;
import org.amdatu.security.authentication.authservice.AuthenticationService;
import org.amdatu.security.authentication.authservice.RejectInfo;
import org.amdatu.security.tokenprovider.TokenProvider;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

/**
 * Test cases for {@link AuthenticationHandler}.
 */
public class AuthenticationResourceTest {
    // Managed by Felix DM...
    private volatile AuthenticationHandler m_authHandler;
    private volatile TokenProvider m_tokenProvider;
    // Locally managed...
    private volatile TestAuthService m_authService;

    private static Map<String, String> asMap(String... entries) {
        Map<String, String> result = new HashMap<>();
        for (int i = 0; i < entries.length; i += 2) {
            result.put(entries[i], entries[i + 1]);
        }
        return result;
    }

    @Before
    public void setUp() {
        m_authService = spy(new TestAuthService(KNOWN_USER));

        configure(this)
            .add(createComponent()
                .setInterface(AuthenticationService.class.getName(), null)
                .setImplementation(m_authService))
            .add(createServiceDependency()
                .setService(AuthenticationHandler.class).setRequired(true))
            .add(createServiceDependency()
                .setService(TokenProvider.class).setRequired(true))
            .apply();
    }

    @After
    public void tearDown() {
        cleanUp(this);
    }

    @Test
    public void testHandleSessionWithExpiredTokenOk() throws Exception {
        // Create an expired token...
        Map<String, String> tokenAttrs = generateToken(Instant.now().minusSeconds(20), Instant.now().minusSeconds(1));

        String token = m_tokenProvider.generateToken(tokenAttrs);

        assertTrue(m_authService.isLoggedIn(tokenAttrs));

        // verify that when the token is given to m_authHandler that we should be logged out...
        RejectInfo rejectInfo = callAuthHandler_HandleSecurity(token);
        assertEquals(CODE_LOGIN_REQUIRED, rejectInfo.getStatusCode());

        assertFalse(m_authService.isLoggedIn(tokenAttrs));

        verify(m_authService, times(1)).logout(anyMapOf(String.class, String.class));
    }

    @Test
    public void testHandleSessionWithNonExpiredTokenOk() throws Exception {
        // Create a non-expired token...
        Map<String, String> tokenAttrs = generateToken(Instant.now().minusSeconds(10), Instant.now().plusSeconds(10));

        String token = m_tokenProvider.generateToken(tokenAttrs);

        assertTrue(m_authService.isLoggedIn(tokenAttrs));

        // verify that when the token is given to m_authHandler that we should be logged out...
        RejectInfo rejectInfo = callAuthHandler_HandleSecurity(token);
        assertNull(rejectInfo);

        assertTrue(m_authService.isLoggedIn(tokenAttrs));

        verify(m_authService, times(0)).logout(anyMapOf(String.class, String.class));
    }

    @Test
    public void testHandleSessionWithoutTokenOk() throws Exception {
        RejectInfo rejectInfo = callAuthHandler_HandleSecurity(null);
        assertEquals(CODE_INVALID_REQUEST, rejectInfo.getStatusCode());
    }

    private RejectInfo callAuthHandler_HandleSecurity(String token) throws IOException {
        HttpServletRequest req = createMockServletRequest(token);
        HttpServletResponse resp = mock(HttpServletResponse.class);

        // verify that when the token is given to m_authHandler that we should be logged out...
        Optional<RejectInfo> rejectInfo = m_authHandler.handleSecurity(req, resp);
        return rejectInfo.orElse(null);
    }

    private HttpServletRequest createMockServletRequest(String token) {
        final Map<String, Object> attrs = new HashMap<>();

        HttpServletRequest req = mock(HttpServletRequest.class);
        doAnswer(mock -> {
            attrs.put(mock.getArguments()[0].toString(), mock.getArguments()[1]);
            return null;
        }).when(req).setAttribute(anyString(), anyObject());

        when(req.getAttribute(anyString())).thenAnswer(mock -> {
            return attrs.get(mock.getArguments()[0]);
        });
        if (token != null) {
            when(req.getCookies()).thenReturn(new Cookie[] { new Cookie("amdatu_token", token) });
        }

        return req;
    }

    private Map<String, String> generateToken(Instant notBefore, Instant expiration) {
        String sessionId = m_authService.generateSessionID();
        String nbf = Long.toString(notBefore.getEpochSecond());
        String exp = Long.toString(expiration.getEpochSecond());

        // Mock the case where the token has already been stored...
        m_authService.injectSession(KNOWN_USER, sessionId);

        Map<String, String> attrs = asMap(SUBJECT, KNOWN_USER,
            PROVIDER_TYPE, "itest",
            NOT_BEFORE, nbf, 
            EXPIRATION_TIME, exp, 
            SESSION_ID, sessionId);

        // Make sure our preconditions are met...
        assertTrue(m_authService.isLoggedIn(attrs));

        return attrs;
    }
}
