/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.test.authservice;

import static java.util.Arrays.asList;
import static org.amdatu.security.authentication.idprovider.IdProviderConstants.PROVIDER_TYPE;
import static org.amdatu.security.tokenprovider.TokenConstants.SUBJECT;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;

import org.amdatu.security.account.Account;
import org.amdatu.security.account.Account.State;
import org.amdatu.security.account.AccountAdminBackend;
import org.amdatu.security.account.AccountExistsException;
import org.amdatu.security.authentication.authservice.AuthenticationException;
import org.amdatu.security.authentication.authservice.AuthenticationService;
import org.amdatu.security.authentication.idprovider.UserLookupService;

/**
 * Mock implementation of {@link AuthenticationService}.
 */
public class TestAuthService implements AuthenticationService, UserLookupService, AccountAdminBackend {
    public static final String KNOWN_USER = "known_user@localhost";

    public static final String SESSION_ID = "sessionID";
    public static final String ERR_ACCESS_DENIED = "access_denied";

    private final Set<String> m_credentialKeys;
    private final Set<String> m_knownUserIds;
    private final Map<String, Account> m_accounts;
    private final Map<String, String> m_sessions;

    /**
     * Creates a new {@link TestAuthService} instance.
     */
    public TestAuthService(String... knownUserIds) {
        m_credentialKeys = new HashSet<>(asList("email", "upn"));
        m_knownUserIds = new HashSet<>(asList(knownUserIds));
        m_accounts = new HashMap<>();
        m_sessions = new HashMap<>();

        Map<String, String> creds = new HashMap<>();
        creds.put("email", KNOWN_USER);
        creds.put("password", "secret");
        m_accounts.put(KNOWN_USER, new Account(KNOWN_USER, creds, State.NORMAL, null, false));
    }

    @Override
    public boolean accountExists(String accountId) {
        return m_accounts.containsKey(accountId);
    }

    @Override
    public void create(Account account) throws AccountExistsException {
        if (m_accounts.containsKey(account.getId())) {
            throw new AccountExistsException();
        }
        m_accounts.putIfAbsent(account.getId(), account);
    }

    public String generateSessionID() {
        return UUID.randomUUID().toString();
    }

    @Override
    public Optional<Account> get(String accountId) {
        return Optional.ofNullable(m_accounts.get(accountId));
    }

    @Override
    public Optional<String> getUserId(Map<String, String> credentials) {
        for (String credentialKey : m_credentialKeys) {
            String userName = credentials.get(credentialKey);
            if (m_knownUserIds.contains(userName)) {
                return Optional.of(userName);
            }
        }

        return Optional.empty();
    }

    public void injectSession(String subject, String sessionID) {
        m_sessions.put(subject, sessionID);
    }

    @Override
    public boolean isLoggedIn(Map<String, String> credentials) {
        String subject = credentials.get(SUBJECT);
        String sessionID = credentials.get(SESSION_ID);
        String storedSessionID = m_sessions.get(subject);

        return storedSessionID != null && storedSessionID.equals(sessionID);
    }

    @Override
    public void login(Map<String, String> credentials, Map<String, String> context) {
        String provider = credentials.get(PROVIDER_TYPE);
        if (provider == null) {
            System.out.printf("[WARN] Login failed for %s: invalid provider type?!%n", credentials);
            throw new AuthenticationException("invalid_provider");
        }

        String subject = credentials.get(SUBJECT);
        String sessionID = credentials.get(SESSION_ID);
        String storedSessionID = m_sessions.get(subject);

        if ((storedSessionID == null && sessionID != null) || (storedSessionID != null && !storedSessionID.equals(sessionID))) {
            System.out.printf("[WARN] Login failed for %s: invalid session ID. Expected: %s, but got: %s.%n", credentials, storedSessionID, sessionID);
            throw new AuthenticationException(ERR_ACCESS_DENIED);
        }

        if (sessionID == null && storedSessionID == null) {
            // Generate a new session ID...
            sessionID = generateSessionID();

            credentials.put(SESSION_ID, sessionID);
            // Store for later use...
            m_sessions.put(subject, sessionID);
        }

        System.out.printf("[INFO] Login successful for %s! Credentials were %s.%n", subject, credentials);
    }

    @Override
    public void logout(Map<String, String> credentials) {
        if (credentials.isEmpty()) {
            // Nothing to do...
            return;
        }

        String subject = credentials.get(SUBJECT);
        String sessionID = credentials.get(SESSION_ID);

        if (!m_sessions.remove(subject, sessionID)) {
            System.out.printf("[WARN] Logout failed for %s: invalid session ID. Expected: %s, but got: %s.%n", credentials, sessionID, m_sessions.get(subject));
            throw new AuthenticationException(ERR_ACCESS_DENIED);
        }

        System.out.printf("[INFO] Logout successful for %s! Credentials were %s.%n", subject, credentials);
    }

    @Override
    public void update(Account account) {
        m_accounts.put(account.getId(), account);
    }
}
