/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.authentication.test.idprovider;

import static org.amdatu.testing.configurator.TestConfigurator.cleanUp;
import static org.amdatu.testing.configurator.TestConfigurator.configure;
import static org.amdatu.testing.configurator.TestConfigurator.createComponent;
import static org.amdatu.testing.configurator.TestConfigurator.createConfiguration;
import static org.amdatu.testing.http.HttpTestConfigurator.createWaitForHttpEndpoints;
import static org.amdatu.testing.http.HttpTestConfigurator.useLocalCookieJar;
import static org.amdatu.web.rest.jaxrs.AmdatuWebRestConstants.*;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.TimeUnit;

import org.amdatu.security.account.AccountAdminBackend;
import org.amdatu.security.authentication.authservice.AuthenticationService;
import org.amdatu.security.authentication.idprovider.UserLookupService;
import org.amdatu.security.authentication.test.app.MyAuthService;
import org.amdatu.security.authentication.test.app.MyLocalApp;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.github.kevinsawicki.http.HttpRequest;

/**
 * Test cases for LocalIdProvider.
 */
public class LocalIdProviderTest {
    public static final String[] KNOWN_USERS = { "known_user@localhost" };

    private final MyAuthService m_authService = new MyAuthService(KNOWN_USERS);
    private final MyLocalApp m_app = new MyLocalApp();
    
    private String m_baseURI;

    @Before
    public void setUp() {
        Properties appProps = new Properties();
        appProps.put(JAX_RS_RESOURCE_BASE, "/");

        int port = Integer.getInteger("org.osgi.service.http.port", 8080);
        m_baseURI = String.format("http://localhost:%d/", port);

        configure(this)
            .add(createComponent()
                .setInterface(Object.class.getName(), appProps)
                .setImplementation(m_app))
            .add(createComponent()
                .setInterface(new String[] { AuthenticationService.class.getName(), AccountAdminBackend.class.getName() }, null)
                .setImplementation(m_authService))
            .add(createComponent()
                .setInterface(UserLookupService.class.getName(), null)
                .setImplementation(m_authService))
            .add(createConfiguration("org.amdatu.security.account.admin")
                .set("keyAccountId", "email"))
            .add(createConfiguration("org.amdatu.security.authentication.idprovider.local")
                .set("credentialKeys", "email, password",
                    "keyUri", "random:32",
                    "timestepWindow", "1"))
            .add(createConfiguration("org.amdatu.security.authentication.authcontext")
                .set("osgi.http.whiteboard.context.path", "/authtest"))
            .add(createConfiguration("org.amdatu.security.authentication.authservice")
                .set("preLoginURL", m_baseURI + "authtest/index.html",
                    "landingPageURL", m_baseURI + "my/app",
                    "failureURL", m_baseURI + "my/failure"))
            .add(createWaitForHttpEndpoints(m_baseURI + "authtest/rest/login/providers"))
            .add(useLocalCookieJar())
            .apply();
    }

    @After
    public void tearDown() {
        cleanUp(this);
    }

    @Test
    public void testUseLocalIdProviderOk() throws Exception {
        Map<String, String> form = new HashMap<>();
        form.put("email", KNOWN_USERS[0]);
        form.put("password", "secret");
        form.put("providerType", "local");

        HttpRequest loginPost = HttpRequest.post(m_baseURI + "authtest/rest/login").form(form).followRedirects(true);
        assertEquals(200, loginPost.code());

        assertTrue(m_app.m_appCalled.await(5, TimeUnit.SECONDS));
        assertFalse(m_app.m_failureCalled.await(1, TimeUnit.MILLISECONDS));

        assertTrue(m_authService.m_getUserIdCalled.await(1, TimeUnit.MILLISECONDS));
        assertTrue(m_authService.m_loginCalled.await(1, TimeUnit.MILLISECONDS));
        assertFalse(m_authService.m_logoutCalled.await(1, TimeUnit.MILLISECONDS));
    }

    @Test
    public void testUseLocalIdProvider_UnknownUser() throws Exception {
        Map<String, String> form = new HashMap<>();
        form.put("email", "unknown@localhost");
        form.put("password", "does_not_matter");
        form.put("providerType", "local");

        HttpRequest loginPost = HttpRequest.post(m_baseURI + "authtest/rest/login").form(form).followRedirects(true);
        assertEquals(200, loginPost.code());

        assertTrue(m_app.m_failureCalled.await(5, TimeUnit.SECONDS));
        assertFalse(m_app.m_appCalled.await(1, TimeUnit.MILLISECONDS));

        assertFalse(m_authService.m_getUserIdCalled.await(1, TimeUnit.MILLISECONDS));
        assertFalse(m_authService.m_loginCalled.await(1, TimeUnit.MILLISECONDS));
        assertFalse(m_authService.m_logoutCalled.await(1, TimeUnit.MILLISECONDS));
    }

    @Test
    public void testUseLocalIdProviderWithExpiredAuthURL() throws Exception {
        Map<String, String> form = new HashMap<>();
        form.put("email", KNOWN_USERS[0]);
        form.put("password", "secret");
        form.put("providerType", "local");

        HttpRequest loginPost = HttpRequest.post(m_baseURI + "authtest/rest/login").form(form).followRedirects(false);
        assertEquals(303, loginPost.code());

        HttpRequest redirectGet = HttpRequest.get(loginPost.location());

        TimeUnit.SECONDS.sleep(2L); // should invalidate our code...

        assertEquals(200, redirectGet.code());

        assertTrue(m_app.m_failureCalled.await(5, TimeUnit.SECONDS));
        assertFalse(m_app.m_appCalled.await(1, TimeUnit.MILLISECONDS));

        assertFalse(m_authService.m_getUserIdCalled.await(1, TimeUnit.MILLISECONDS));
        assertFalse(m_authService.m_loginCalled.await(1, TimeUnit.MILLISECONDS));
        assertFalse(m_authService.m_logoutCalled.await(1, TimeUnit.MILLISECONDS));
    }
}
