/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.password.util;

import static org.junit.Assert.assertArrayEquals;

import org.amdatu.security.password.util.Hkdf;
import org.junit.Test;

/**
 * Test cases for {@link Hkdf}.
 */
public class HkdfTest {

    private static final String HMAC_SHA1 = "HmacSHA1";
    private static final String HMAC_SHA256 = "HmacSHA256";

    /**
     * See RFC 5869, appendix A.1.
     */
    @Test
    public void testCase1() throws Exception {
        byte[] ikm = asBytes("0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b");
        byte[] salt = asBytes("000102030405060708090a0b0c");
        byte[] info = asBytes("f0f1f2f3f4f5f6f7f8f9");
        int l = 42;

        Hkdf hkdf = new Hkdf.Builder().setAlgorithm(HMAC_SHA256).setInitialKeyingMaterial(ikm).setSalt(salt).build();
        byte[] actual = hkdf.deriveKey(info, l);

        byte[] okm = asBytes("3cb25f25faacd57a90434f64d0362f2a2d2d0a90cf1a5a4c5db02d56ecc4c5bf34007208d5b887185865");

        assertArrayEquals(okm, actual);
    }

    /**
     * See RFC 5869, appendix A.2.
     */
    @Test
    public void testCase2() throws Exception {
        byte[] ikm = asBytes("000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f"
            + "202122232425262728292a2b2c2d2e2f303132333435363738393a3b3c3d3e3f404142434445464748494a4b4c4d4e4f");
        byte[] salt = asBytes("606162636465666768696a6b6c6d6e6f707172737475767778797a7b7c7d7e7f"
            + "808182838485868788898a8b8c8d8e8f909192939495969798999a9b9c9d9e9fa0a1a2a3a4a5a6a7a8a9aaabacadaeaf");
        byte[] info = asBytes("b0b1b2b3b4b5b6b7b8b9babbbcbdbebfc0c1c2c3c4c5c6c7c8c9cacbcccdcecf"
            + "d0d1d2d3d4d5d6d7d8d9dadbdcdddedfe0e1e2e3e4e5e6e7e8e9eaebecedeeeff0f1f2f3f4f5f6f7f8f9fafbfcfdfeff");
        int l = 82;

        Hkdf hkdf = new Hkdf.Builder().setAlgorithm(HMAC_SHA256).setInitialKeyingMaterial(ikm).setSalt(salt).build();
        byte[] actual = hkdf.deriveKey(info, l);

        byte[] okm = asBytes("b11e398dc80327a1c8e7f78c596a49344f012eda2d4efad8a050cc4c19afa97c5"
            + "9045a99cac7827271cb41c65e590e09da3275600c2f09b8367793a9aca3db71cc30c58179ec3e87c14c01d5c1f3434f1d87");

        assertArrayEquals(okm, actual);
    }

    /**
     * See RFC 5869, appendix A.3.
     */
    @Test
    public void testCase3() throws Exception {
        byte[] ikm = asBytes("0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b");
        byte[] salt = asBytes("");
        byte[] info = asBytes("");
        int l = 42;

        Hkdf hkdf = new Hkdf.Builder().setAlgorithm(HMAC_SHA256).setInitialKeyingMaterial(ikm).setSalt(salt).build();
        byte[] actual = hkdf.deriveKey(info, l);

        byte[] okm = asBytes("8da4e775a563c18f715f802a063c5a31b8a11f5c5ee1879ec3454e5f3c738d2d9d201395faa4b61a96c8");

        assertArrayEquals(okm, actual);
    }

    /**
     * See RFC 5869, appendix A.4.
     */
    @Test
    public void testCase4() throws Exception {
        byte[] ikm = asBytes("0b0b0b0b0b0b0b0b0b0b0b");
        byte[] salt = asBytes("000102030405060708090a0b0c");
        byte[] info = asBytes("f0f1f2f3f4f5f6f7f8f9");
        int l = 42;

        Hkdf hkdf = new Hkdf.Builder().setAlgorithm(HMAC_SHA1).setInitialKeyingMaterial(ikm).setSalt(salt).build();
        byte[] actual = hkdf.deriveKey(info, l);

        byte[] okm = asBytes("085a01ea1b10f36933068b56efa5ad81a4f14b822f5b091568a9cdd4f155fda2c22e422478d305f3f896");

        assertArrayEquals(okm, actual);
    }

    /**
     * See RFC 5869, appendix A.5.
     */
    @Test
    public void testCase5() throws Exception {
        byte[] ikm = asBytes("000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f202122232425262728292a2b2c2d2e2f303132333435363738393a3b3c3d3e3f404142434445464748494a4b4c4d4e4f");
        byte[] salt = asBytes("606162636465666768696a6b6c6d6e6f707172737475767778797a7b7c7d7e7f808182838485868788898a8b8c8d8e8f909192939495969798999a9b9c9d9e9fa0a1a2a3a4a5a6a7a8a9aaabacadaeaf");
        byte[] info = asBytes("b0b1b2b3b4b5b6b7b8b9babbbcbdbebfc0c1c2c3c4c5c6c7c8c9cacbcccdcecfd0d1d2d3d4d5d6d7d8d9dadbdcdddedfe0e1e2e3e4e5e6e7e8e9eaebecedeeeff0f1f2f3f4f5f6f7f8f9fafbfcfdfeff");
        int l = 82;

        Hkdf hkdf = new Hkdf.Builder().setAlgorithm(HMAC_SHA1).setInitialKeyingMaterial(ikm).setSalt(salt).build();
        byte[] actual = hkdf.deriveKey(info, l);

        byte[] okm = asBytes("0bd770a74d1160f7c9f12cd5912a06ebff6adcae899d92191fe4305673ba2ffe8fa3f1a4e5ad79f3f334b3b202b2173c486ea37ce3d397ed034c7f9dfeb15c5e927336d0441f4c4300e2cff0d0900b52d3b4");

        assertArrayEquals(okm, actual);
    }

    /**
     * See RFC 5869, appendix A.6.
     */
    @Test
    public void testCase6() throws Exception {
        byte[] ikm = asBytes("0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b0b");
        byte[] salt = asBytes("");
        byte[] info = asBytes("");
        int l = 42;

        Hkdf hkdf = new Hkdf.Builder().setAlgorithm(HMAC_SHA1).setInitialKeyingMaterial(ikm).setSalt(salt).build();
        byte[] actual = hkdf.deriveKey(info, l);

        byte[] okm = asBytes("0ac1af7002b3d761d1e55298da9d0506b9ae52057220a306e07b6b87e8df21d0ea00033de03984d34918");

        assertArrayEquals(okm, actual);
    }

    /**
     * See RFC 5869, appendix A.7.
     */
    @Test
    public void testCase7() throws Exception {
        byte[] ikm = asBytes("0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c0c");
        byte[] salt = null;
        byte[] info = null;
        int l = 42;

        Hkdf hkdf = new Hkdf.Builder().setAlgorithm(HMAC_SHA1).setInitialKeyingMaterial(ikm).setSalt(salt).build();
        byte[] actual = hkdf.deriveKey(info, l);

        byte[] okm = asBytes("2c91117204d745f3500d636a62f64f0ab3bae548aa53d423b0d1f27ebba6f5e5673a081d70cce7acfc48");

        assertArrayEquals(okm, actual);
    }

    private static byte[] asBytes(String data) {
        byte[] result = new byte[data.length() / 2];
        for (int x = 0; x < result.length; x++) {
            int start = 2 * x;
            int end = start + 2;
            result[x] = (byte) Integer.parseInt(data.substring(start, end), 16);
        }
        return result;
    }
}
